#include <stdarg.h>
#include <stdlib.h>

#define UPNP_HAVE_TOOLS 1
#define UPNP_HAVE_DEBUG 0
#include <upnp.h>
#include <ithread.h>
#include <upnptools.h>

#include "UPnP_Util.h"

//#define UpnpUtil_Print printf
//#define UpnpUtil_Print(...)

/********************************************************************************
 * UpnpUtil_GetElementValue
 *
 * Description: 
 *       Given a DOM node such as <Channel>11</Channel>, this routine
 *       extracts the value (e.g., 11) from the node and returns it as 
 *       a string. The string must be freed by the caller using 
 *       free.
 *
 * Parameters:
 *   node -- The DOM node from which to extract the value
 *
 ********************************************************************************/

char *
UpnpUtil_GetElementValue( IN IXML_Element * element )
{

    IXML_Node *child = ixmlNode_getFirstChild( ( IXML_Node * ) element );

    char *temp = NULL;

    if( ( child != 0 ) && ( ixmlNode_getNodeType( child ) == eTEXT_NODE ) ) {
        temp = strdup( ixmlNode_getNodeValue( child ) );
    }

    return temp;
}

/********************************************************************************
 * UpnpUtil_GetFirstServiceList
 *
 * Description: 
 *       Given a DOM node representing a UPnP Device Description Document,
 *       this routine parses the document and finds the first service list
 *       (i.e., the service list for the root device).  The service list
 *       is returned as a DOM node list.
 *
 * Parameters:
 *   node -- The DOM node from which to extract the service list
 *
 ********************************************************************************/
IXML_NodeList *
UpnpUtil_GetFirstServiceList( IN IXML_Document * doc )
{
    IXML_NodeList *ServiceList = NULL;
    IXML_NodeList *servlistnodelist = NULL;
    IXML_Node *servlistnode = NULL;

    servlistnodelist =
        ixmlDocument_getElementsByTagName( doc, "serviceList" );
    if( servlistnodelist && ixmlNodeList_length( servlistnodelist ) ) {

        /*
           we only care about the first service list, from the root device 
         */
        servlistnode = ixmlNodeList_item( servlistnodelist, 0 );

        /*
           create as list of DOM nodes 
         */
        ServiceList =
            ixmlElement_getElementsByTagName( ( IXML_Element * )
                                              servlistnode, "service" );
    }

    if( servlistnodelist )
        ixmlNodeList_free( servlistnodelist );

    return ServiceList;
}

/********************************************************************************
 * UpnpUtil_GetFirstDocumentItem
 *
 * Description: 
 *       Given a document node, this routine searches for the first element
 *       named by the input string item, and returns its value as a string.
 *       String must be freed by caller using free.
 * Parameters:
 *   doc -- The DOM document from which to extract the value
 *   item -- The item to search for
 *
 ********************************************************************************/
char *
UpnpUtil_GetFirstDocumentItem( IN IXML_Document * doc,
                                 IN const char *item )
{
    IXML_NodeList *nodeList = NULL;
    IXML_Node *textNode = NULL;
    IXML_Node *tmpNode = NULL;

    char *ret = NULL;

    nodeList = ixmlDocument_getElementsByTagName( doc, ( char * )item );

    if( nodeList ) {
        if( ( tmpNode = ixmlNodeList_item( nodeList, 0 ) ) ) {
            textNode = ixmlNode_getFirstChild( tmpNode );

            ret = strdup( ixmlNode_getNodeValue( textNode ) );
        }
    }

    if( nodeList )
        ixmlNodeList_free( nodeList );
    return ret;
}

/********************************************************************************
 * UpnpUtil_GetFirstElementItem
 *
 * Description: 
 *       Given a DOM element, this routine searches for the first element
 *       named by the input string item, and returns its value as a string.
 *       The string must be freed using free.
 * Parameters:
 *   node -- The DOM element from which to extract the value
 *   item -- The item to search for
 *
 ********************************************************************************/
char *
UpnpUtil_GetFirstElementItem( IN IXML_Element * element,
                                IN const char *item )
{
    IXML_NodeList *nodeList = NULL;
    IXML_Node *textNode = NULL;
    IXML_Node *tmpNode = NULL;

    char *ret = NULL;

    nodeList = ixmlElement_getElementsByTagName( element, ( char * )item );

    if( nodeList == NULL ) {
        //UpnpUtil_Print( "Error finding %s in XML Node\n", item );
        return NULL;
    }

    if( ( tmpNode = ixmlNodeList_item( nodeList, 0 ) ) == NULL ) {
        //UpnpUtil_Print( "Error finding %s value in XML Node\n", item );
        ixmlNodeList_free( nodeList );
        return NULL;
    }

    textNode = ixmlNode_getFirstChild( tmpNode );

    ret = strdup( ixmlNode_getNodeValue( textNode ) );

    if( !ret ) {
        //UpnpUtil_Print( "Error allocating memory for %s in XML Node\n",item );
        ixmlNodeList_free( nodeList );
        return NULL;
    }

    ixmlNodeList_free( nodeList );

    return ret;
}
/********************************************************************************
 * UpnpUtil_FindAndParseService
 *
 * Description: 
 *       This routine finds the first occurance of a service in a DOM representation
 *       of a description document and parses it.  
 *
 * Parameters:
 *   DescDoc -- The DOM description document
 *   location -- The location of the description document
 *   serviceSearchType -- The type of service to search for
 *   serviceId -- OUT -- The service ID
 *   eventURL -- OUT -- The event URL for the service
 *   controlURL -- OUT -- The control URL for the service
 *
 ********************************************************************************/
int
UpnpUtil_FindAndParseService( IN IXML_Document * DescDoc,
                                IN char *location,
                                IN char *serviceType,
                                OUT char **serviceId,
                                OUT char **eventURL,
                                OUT char **controlURL )
{
    int i,
      length,
      found = 0;
    int ret;
    char *tempServiceType = NULL;
    char *baseURL = NULL;
    char *base;
    char *relcontrolURL = NULL,
     *releventURL = NULL;
    IXML_NodeList *serviceList = NULL;
    IXML_Element *service = NULL;

    baseURL = UpnpUtil_GetFirstDocumentItem( DescDoc, "URLBase" );

    if( baseURL )
        base = baseURL;
    else
        base = location;

    serviceList = UpnpUtil_GetFirstServiceList( DescDoc );
    length = ixmlNodeList_length( serviceList );
    for( i = 0; i < length; i++ ) {
        service = ( IXML_Element * ) ixmlNodeList_item( serviceList, i );
        tempServiceType =
            UpnpUtil_GetFirstElementItem( ( IXML_Element * ) service,
                                            "serviceType" );

        if( strncmp( tempServiceType, serviceType, strlen(serviceType) ) == 0 ) {
            //UpnpUtil_Print( "Found service: %s\n", serviceType );
            *serviceId =
                UpnpUtil_GetFirstElementItem( service, "serviceId" );
            //UpnpUtil_Print( "serviceId: %s\n", ( *serviceId ) );
            relcontrolURL =
                UpnpUtil_GetFirstElementItem( service, "controlURL" );
            releventURL =
                UpnpUtil_GetFirstElementItem( service, "eventSubURL" );

            *controlURL =
                (char *)malloc( strlen( base ) + strlen( relcontrolURL ) + 1 );
            if( *controlURL ) {
                ret = UpnpResolveURL( base, relcontrolURL, *controlURL );
                //if( ret != UPNP_E_SUCCESS )
                    //UpnpUtil_Print( "Error generating controlURL from %s + %s\n",base, relcontrolURL );
            }

            *eventURL =
                (char *)malloc( strlen( base ) + strlen( releventURL ) + 1 );
            if( *eventURL ) {
                ret = UpnpResolveURL( base, releventURL, *eventURL );
                //if( ret != UPNP_E_SUCCESS )
                    //UpnpUtil_Print( "Error generating eventURL from %s + %s\n", base,releventURL );
            }

            if( relcontrolURL )
                free( relcontrolURL );
            if( releventURL )
                free( releventURL );
            relcontrolURL = releventURL = NULL;

            found = 1;
            break;
        }

        if( tempServiceType )
            free( tempServiceType );
        tempServiceType = NULL;
    }

    if( tempServiceType )
        free( tempServiceType );
    if( serviceList )
        ixmlNodeList_free( serviceList );
    if( baseURL )
        free( baseURL );

    return ( found );
}

char *UpnpUtil_UnescapeXML(const char *inBuf)
{
    if (!inBuf) return NULL;

    char *result = (char *)malloc(strlen(inBuf) + 1);
    if (!result)return NULL;

    char *outBuf = result;

    while(*inBuf)
    {
        if (*inBuf == '&')
        {
            if (strncmp("&amp;", inBuf, 5) == 0)
            {
                *outBuf = '&';
                inBuf += 5;
                outBuf ++;
            }
            else if (strncmp("&quot;", inBuf, 6) == 0)
            {
                *outBuf = '"';
                inBuf += 6;
                outBuf ++;
            }
            else if (strncmp("&gt;", inBuf, 4) == 0)
            {
                *outBuf = '>';
                inBuf += 4;
                outBuf ++;
            }
            else if (strncmp("&lt;", inBuf, 4) == 0)
            {
                *outBuf = '<';
                inBuf += 4;
                outBuf ++;
            }
            else *outBuf++ = *inBuf++;
        }
        else
        {
            *outBuf++ = *inBuf++;
        }
    }
    *outBuf = '\0';
    return result;
};

/*
char *strdup(const char *s)
{
    char *ret = (char *)malloc(strlen(s) + 1);
    strcpy(ret, s);
    return ret;
};*/
/*
int strcasecmp(const char *s1, const char *s2)
{
    const char CharMap[26] = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                              'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
                              'u', 'v', 'w', 'x', 'y', 'z'};
    const char *c1;
    const char *c2;

    while((*s1 != 0) && (*s2 != 0))
    {
        c1 = s1;
        c2 = s2;
        if ((*s1 >= 'A') && (*s1 <= 'Z'))
            c1 = CharMap + (*s1) - 'A';
        if ((*s2 >= 'A') && (*s2 <= 'Z'))
            c2 = CharMap + (*s2) - 'A';

        if (*c1 != *c2)
            return 1;
        s1++;
        s2++;
    }
    if ((*s1 != 0) || (*s2 != 0))
        return 1;
    return 0;
}
*/

const char *xml_getChildElementValue(IXML_Element * parent, const char * tagName)
{
    if (!parent) return NULL;
    if (!tagName) return NULL;

    char * s = strdup(tagName);
    IXML_NodeList * nodeList = ixmlElement_getElementsByTagName(parent, s);
    free(s);
    if (!nodeList) return NULL;

    IXML_Node * element = ixmlNodeList_item( nodeList, 0 );
    ixmlNodeList_free( nodeList );
    if ( !element ) return NULL;

    IXML_Node * textNode = ixmlNode_getFirstChild(element);
    if ( !textNode ) return NULL;

    return ixmlNode_getNodeValue(textNode);
}

