#ifndef _CDS_MEDIACLASS_H
#define _CDS_MEDIACLASS_H

/*
 *	The purpose of this module is to abstract the concept of a upnp media class,
 *	(eg, the data stored in the upnp:class DIDL-Lite element) into an unsigned int
 *	form. 
 *
 *	The module also provides the means to allow vendors to add their own custom
 *	media classes to the existing infrastructure. 
 *
 */

/*
 *	A media class is represented in string form for UPnP, but can be exposed to applications
 *	through an unsigned int of 16 bits. If the value of the media class is CDS_CLASS_MASK_BADCLASS (zero), 
 *	then it should mean that the media class is unassigned or invalid in some form.
 *
 *	Described below is a description of how the bits are used in the value.
 *	Please note that the bit-string is organized such that sorting the 16-bit values
 *	will correspond appropriately with the sorting order of the same classes in 
 *	in string form.
 *
 *	Bit:	Description
 *  -----------------------------------------------------------------------------------------------
 *	0-5:	6 bits to represent customized extensions.
 *			If vendors need more extensions, then a 32-bit value can be used.
 *			Extract these bits out using a bitwise-AND operation of media class & CDS_CLASS_MASK_MINOR2,
 *			shift right the resulting value by CDS_SHIFT_MINOR2_TYPE bits,
 *			and use the value as an index into the as CDS_CLASS_MINOR2_TYPE array.
 *
 *	6-9:	4 bits to represent minor classes
 *			(12 CDS-normative values: musicAlbum, photoAlbum, musicGenre, movieGenre, 
 *			musicArtist, audioBook, audioBroadcast, musicTrack, photo, movie, 
 *			musicVideoClip, videoBroadcast)
 *			Extract these bits out using a bitwise-AND operation of media class & CDS_CLASS_MASK_MINOR1,
 *			shift right the resulting value by CDS_SHIFT_MINOR1_TYPE bits,
 *			and use the value as an index into the as CDS_CLASS_MINOR1_TYPE array.
 *	
 *	10-13:	4 bits to represent major classes
 *			(12 CDS-normative values: album, genre, person, playlistContainer, 
 *			storageFolder, storageSystem, storageVolume, audioItem, imageItem, 
 *			playlistItem, textItem, videoItem)
 *			Extract these bits out using a bitwise-AND operation of media class & CDS_CLASS_MASK_MAJOR,
 *			shift right the resulting value by CDS_SHIFT_MAJOR_TYPE bits,
 *			and use the value as an index into the as CDS_CLASS_MAJOR_TYPE array.
 *
 *	14:		Flag: object is container. 
 *	15:		Flag: object is item. 
 *
 *			Notes about bits 14-15.
 *			Extract these bits out using a bitwise-AND operation of media class & CDS_CLASS_MASK_OBJECT_TYPE,
 *			shift right the resulting value by CDS_SHIFT_OBJECT_TYPE bits,
 *			and use the value as an index into the as CDS_CLASS_OBJECT_TYPE array.
 *
 *	For all intents and purposes, if a bitwise-AND operation with a CDS_CLASS_MASK_xxx value yields
 *	a value of zero, then it means that the particulate portion of the media class remains unassigned.
 *
 *	To obtain a string representation of a media class, concatenate the strings that represent the
 *	object-type, major-type, minor type, and minor2 types with each substring separated with a 
 *	a period (.) character. There is no trailing dot character. For example, the
 *	media class of 0xA200 (which has binary form: 10 1000 1000 000000) would have a 
 *	corresponding string representation of "object.item.audioItem.musicTrack" because:
 *		CDS_CLASS_OBJECT_TYPE[(0xA200 & CDS_CLASS_MASK_OBJECTTYPE) >> CDS_SHIFT_OBJECT_TYPE] == "object.item"
 *		CDS_CLASS_MAJOR_TYPE[(0xA200 & CDS_CLASS_MASK_MAJOR) >> CDS_SHIFT_MAJOR_TYPE] == "audioItem"
 *		CDS_CLASS_MINOR1_TYPE[(0xA200 & CDS_CLASS_MASK_MAJOR) >> CDS_SHIFT_MINOR1_TYPE] == "musicTrack"
 *
 *	Another added benefit of this architecture is that it's friendly to boolean operations that involve
 *	the media class. For example, the following are common operations.
 *
 *		((mediaClass & CDS_CLASS_MASK_ITEM))
 *			: Indicates if the media class is some kind of item
 *
 *		((mediaClass & CDS_CLASS_MASK_CONTAINER))			
 *			: Indicates if the media class is some kind of container
 *
 *		((mediaClass & CDS_CLASS_MASK_MAJOR_AUDIOITEM)==CDS_CLASS_MASK_MAJOR_AUDIOITEM)	
 *			: Indicates if the media class is some kind of audio related item
 *
 *	The header file defines a set of CDS_CLASS_MASK_xxx values to match up with the 
 *	normative set of media classes defined in UPnP AV.
 *
 *	To obtain a value for a normative media class, use one of the CDS_MEDIACLASS_xxx values. 
 *	For example, to get the media class value for an audio broadcast, use CDS_MEDIACLASS_AUDIOBROADCAST.
 *	Using CDS_CLASS_MASK_MINOR1_AUDIOBROADCAST will only set the minor1 portion of the media class,
 *	which is an incomplete media class. 
 */




/* Normative set of object types: eg, object.item or object.container */
extern const char* CDS_CLASS_OBJECT_TYPE[];	

/* Normative set of major classes - first sublevel from item or container - ex: object.item.audioItem*/
extern const char* CDS_CLASS_MAJOR_TYPE[];	

/* Normative set of minor class designations - second sublevel from item or container - ex: object.item.audioItem.musicTrack */
extern const char* CDS_CLASS_MINOR1_TYPE[];	

/* Custom set of minor designations - third sublevel from item or container. */
extern const char* CDS_CLASS_MINOR2_TYPE[];	

/* The number of strings in CDS_CLASS_OBJECT_TYPE */
#define CDS_CLASS_OBJECT_TYPE_LEN	3

/* The number of strings in CDS_CLASS_MAJOR_TYPE */
#define CDS_CLASS_MAJOR_TYPE_LEN	13 

/* The number of strings in CDS_CLASS_MINOR1_TYPE */
#define CDS_CLASS_MINOR1_TYPE_LEN	13

/* The number of string in CDS_CLASS_MINOR2_TYPE */
#define CDS_CLASS_MINOR2_TYPE_LEN	1

/* MINOR2 + MINOR1 + MAJOR = 6 + 4 + 4 = 14*/
#define CDS_SHIFT_OBJECT_TYPE					14

/* MINOR2 + MINOR1 = 6 + 4 = 10*/
#define CDS_SHIFT_MAJOR_TYPE					10

/* MINOR2 = 6 */
#define CDS_SHIFT_MINOR1_TYPE					6

/* 0 */
#define CDS_SHIFT_MINOR2_TYPE					0

/*
 *	Use of class masks are described above.
 */

#define CDS_CLASS_MASK_OBJECT_TYPE				0xC000	/* 11 0000 0000 000000 */
#define CDS_CLASS_MASK_MAJOR					0x3C00	/* 00 1111 0000 000000 */
#define CDS_CLASS_MASK_MINOR1					0x03C0	/* 00 0000 1111 000000 */
#define CDS_CLASS_MASK_MINOR2					0x003F	/* 00 0000 0000 111111 */

#define CDS_CLASS_MASK_BADCLASS					0x0000
#define CDS_CLASS_MASK_CONTAINER				0x4000	/* 01 0000 0000 000000 */
#define CDS_CLASS_MASK_ITEM						0x8000	/* 10 0000 0000 000000 */

/* major container classes */
#define CDS_CLASS_MASK_MAJOR_ALBUM				0x0400	/* 00 0001 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_GENRE				0x0800	/* 00 0010 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_PERSON				0x0C00	/* 00 0011 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_PLAYLISTCONTAINER	0x1000	/* 00 0100 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_STORAGEFOLDER		0x1400	/* 00 0101 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_STORAGESYSTEM		0x1800	/* 00 0110 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_STORAGEVOLUME		0x1C00	/* 00 0111 0000 000000 */

/* major item classes */
#define CDS_CLASS_MASK_MAJOR_AUDIOITEM			0x2000	/* 00 1000 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_IMAGEITEM			0x2400	/* 00 1001 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_PLAYLISTITEM		0x2800	/* 00 1010 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_TEXTITEM			0x2C00	/* 00 1011 0000 000000 */
#define CDS_CLASS_MASK_MAJOR_VIDEOITEM			0x3000	/* 00 1100 0000 000000 */


/* minor1 container classes */
#define CDS_CLASS_MASK_MINOR1_MUSICALBUM		0x0040	/* 00 0000 0001 000000 */
#define CDS_CLASS_MASK_MINOR1_PHOTOALBUM		0x0080	/* 00 0000 0010 000000 */
#define CDS_CLASS_MASK_MINOR1_MOVIEGENRE		0x00C0	/* 00 0000 0011 000000 */
#define CDS_CLASS_MASK_MINOR1_MUSICGENRE		0x0100	/* 00 0000 0100 000000 */
#define CDS_CLASS_MASK_MINOR1_MUSICARTIST		0x0140	/* 00 0000 0101 000000 */

/* minor1 item classes */
#define CDS_CLASS_MASK_MINOR1_AUDIOBOOK			0x0180	/* 00 0000 0110 000000 */
#define CDS_CLASS_MASK_MINOR1_AUDIOBROADCAST	0x01C0	/* 00 0000 0111 000000 */
#define CDS_CLASS_MASK_MINOR1_MUSICTRACK		0x0200	/* 00 0000 1000 000000 */
#define CDS_CLASS_MASK_MINOR1_PHOTO				0x0240	/* 00 0000 1001 000000 */
#define CDS_CLASS_MASK_MINOR1_MOVIE				0x0280	/* 00 0000 1010 000000 */
#define CDS_CLASS_MASK_MINOR1_MUSICVIDEOCLIP	0x02C0	/* 00 0000 1011 000000 */
#define CDS_CLASS_MASK_MINOR1_VIDEOBROADCAST	0x0300	/* 00 0000 1100 000000 */

/*
 *	If the application needs to get a mediaClass value for a specific type of
 *	media object, it should use these macros.
 */

#define CDS_MEDIACLASS_ITEM						(CDS_CLASS_MASK_ITEM)
#define CDS_MEDIACLASS_CONTAINER				(CDS_CLASS_MASK_CONTAINER)

#define CDS_MEDIACLASS_IMAGEITEM				(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_IMAGEITEM			)
#define CDS_MEDIACLASS_AUDIOITEM				(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_AUDIOITEM			)
#define CDS_MEDIACLASS_VIDEOITEM				(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_VIDEOITEM			)
#define CDS_MEDIACLASS_PLAYLISTITEM				(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_PLAYLISTITEM		)
#define CDS_MEDIACLASS_TEXTITEM					(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_TEXTITEM			)

#define CDS_MEDIACLASS_PERSON					(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_PERSON				)
#define CDS_MEDIACLASS_PLAYLISTCONTAINER		(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_PLAYLISTCONTAINER	)
#define CDS_MEDIACLASS_ALBUM					(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_ALBUM				)
#define CDS_MEDIACLASS_GENRE					(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_GENRE				)
#define CDS_MEDIACLASS_STORAGESYSTEM			(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_STORAGESYSTEM		)
#define CDS_MEDIACLASS_STORAGEVOLUME			(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_STORAGEVOLUME		)
#define CDS_MEDIACLASS_STORAGEFOLDER			(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_STORAGEFOLDER		)

#define CDS_MEDIACLASS_PHOTO					(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_IMAGEITEM | CDS_CLASS_MASK_MINOR1_PHOTO			)
#define CDS_MEDIACLASS_MUSICTRACK				(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_AUDIOITEM | CDS_CLASS_MASK_MINOR1_MUSICTRACK	)
#define CDS_MEDIACLASS_AUDIOBROADCAST			(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_AUDIOITEM | CDS_CLASS_MASK_MINOR1_AUDIOBROADCAST)
#define CDS_MEDIACLASS_AUDIOBOOK				(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_AUDIOITEM | CDS_CLASS_MASK_MINOR1_AUDIOBOOK		)
#define CDS_MEDIACLASS_MOVIE					(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_VIDEOITEM | CDS_CLASS_MASK_MINOR1_MOVIE			)
#define CDS_MEDIACLASS_VIDEOBROADCAST			(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_VIDEOITEM | CDS_CLASS_MASK_MINOR1_VIDEOBROADCAST)
#define CDS_MEDIACLASS_MUSICVIDEOCLIP			(CDS_CLASS_MASK_ITEM | CDS_CLASS_MASK_MAJOR_VIDEOITEM | CDS_CLASS_MASK_MINOR1_MUSICVIDEOCLIP)

#define CDS_MEDIACLASS_MUSICARTIST				(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_PERSON | CDS_CLASS_MASK_MINOR1_MUSICARTIST	)
#define CDS_MEDIACLASS_MUSICALBUM				(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_ALBUM  | CDS_CLASS_MASK_MINOR1_MUSICALBUM	)
#define CDS_MEDIACLASS_PHOTOALBUM				(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_ALBUM  | CDS_CLASS_MASK_MINOR1_PHOTOALBUM	)
#define CDS_MEDIACLASS_MUSICGENRE				(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_GENRE  | CDS_CLASS_MASK_MINOR1_MUSICGENRE	)
#define CDS_MEDIACLASS_MOVIEGENRE				(CDS_CLASS_MASK_CONTAINER | CDS_CLASS_MASK_MAJOR_GENRE  | CDS_CLASS_MASK_MINOR1_MOVIEGENRE	)

#endif

