#ifndef _FILESYSTEM_ENUMERATOR_H
#define _FILESYSTEM_ENUMERATOR_H

#include "MediaServerAbstraction.h"

enum FSE_ResourceMetadataFlags
{
	FSE_RMF_IsFile				= 0x0001,
	FSE_RMF_IsDirectory			= 0x0002,
	FSE_RMF_IsPlaylist			= 0x0004
};

/*
 *	TODO: You can customize this struct to match the
 *	capabilities of the metadata store. As a note,
 *	there is no requirement for every metadata
 *	store to have the same struct. Although
 *	the FSE returns only a local file path
 *	and a filesize, this information could be sent
 *	to other modules (such as a transcoding engine)
 *	that can provide additional CdsMediaResource
 *	objects for a media object.
 */
struct FSE_ResourceMetadata
{
	int Flags;					/* mapped by (enum FSE_ResourceMetadataFlags) */
	char *LocalFilePath;
	int LocalFilePathLen;
	char *FileExtension;
	int FileExtensionLen;
	unsigned int FileLength;
};


/*
 *	This module provides the interface for enumerating a file system as if it were
 *	a metadata store.
 */

/*
 *	Starts the process for handling a CDS query.
 *
 *	fseObj				: object from FSE_InitFSEState().
 *
 *	cdsQuery			: the details of the CDS metadata query
 *
 *	filter				: a bit string indicating which fields to return;
 *						: bit string is mapped by 'enum CdsFilterBits'
 *
 *	cdsErrorCode		: contains a CDS error code (enum Enum_CdsErrors)
 *						: in case the query cannot be resolved
 *
 *	Returns void* object representing the query, as understood by the FSE.
 */
void* FSE_HandleQuery(void* fseObj, struct MSA_CdsQuery *cdsQuery, unsigned int filter, /*OUT*/ enum Enum_CdsErrors *cdsErrorCode);

/*
 *	Used to obtain metadata for a CdsMediaObject.
 *	This method is called after FSE_HandleQuery() and 
 *	before FSE_HandleOnAcquireUpdateIDAndCleanup().
 *
 *	fseObj				: object from FSE_InitFSEState().
 *
 *	cdsQuery			: the CDS metadata query that was sent to FSE_HandleQuery()
 *
 *	onQueryObj			: the object obtained from FSE_HandleQuery()
 *
 *	filter				: a bit string indicating which fields to return;
 *						: bit string is mapped by 'enum CdsFilterBits'
 *
 *	cdsObj				: allocated CdsMediaObject;
 *						: fields will be initialized to zero
 *
 *	Returns nonzero if metadata was applied to cdsOb. Return value of zero indicates
 *	that there are no more objects to process.
 */
int FSE_HandleProcessObject(void *fseObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, unsigned int filter, /*INOUT*/ struct CdsMediaObject *cdsObj);

int FSE_HandleProcessObjects(void *fseObj, struct MSA_CdsQuery *cdsQuery, void **onQueryObj, unsigned int filter, /*INOUT*/ struct CdsMediaObject **cdsObj, int count);

int FSE_HandleQueryCount(void *fseObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, unsigned int filter);

/*
 *	Used to obtain the 'UpdateID' out parameter value for a CDS response.
 *	This method should at the very end of responding to a CDS query.
 *
 *	fseObj				: object from FSE_InitFSEState().
 *
 *	cdsQuery			: the CDS metadata query that was sent to FSE_HandleQuery()
 *
 *	onQueryObj			: the object obtained from FSE_HandleQuery()
 *
 *	Returns a value to populate the 'UpdateID' parameter in the
 *	CDS response.
 */
unsigned int FSE_HandleOnAcquireUpdateIDAndCleanup(void *fseObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj);

void* FSE_InitFSEState(void* chain, const char *rootDir);

/*
 *	Returns metadata that is useful creating a CdsMediaResource object.
 *	This method is expected to be called after FSE_HandleQuery()
 *	and before FSE_HandleOnAcquireUpdateIDAndCleanup(). Generally,
 *	for every call to FSE_HandleProcessObject() there should be a call
 *	to FSE_GetResourceMetadata().
 *
 *	fseObj				: object from FSE_InitFSEState().
 *
 *	onQueryObj			: the object obtained from FSE_HandleQuery()
 *
 *	res					: allocated FSE_ResourceMetadata; struct will be memset to zero
 *						: Caller needs to call FREE() on char* fields to free the 
 *						: allocated memory.
 */
void FSE_GetResourceMetadata(void *fseObj, void *onQueryObj, /*INOUT*/ struct FSE_ResourceMetadata *res);

//try it by ty
//int TestProcessObject(void *fseObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, unsigned int filter, /*INOUT*/ struct CdsMediaObject *cdsObj);
#endif