#ifndef _RESOURCE_PROVIDER_LOCALFILEHTTP_H
#define _RESOURCE_PROVIDER_LOCALFILEHTTP_H

#include "CdsMediaObject.h"
#include "TransferState.h"

/*
 *	This module is safe to use with the MSL.
 *	It does not use the ->PreSelect or ->PostSelect fields
 *	on links in a thread-chain.
 */

/*
 *	Internally, we have an array of strings
 *	for the base directories where content
 *	is available. 
 */
#define RP_MAX_START_PATHS 255

/*
 *	This creates a state object for the LocalFileHttp ResourceProvider.
 *
 *	chain		: object from ILibCreateChain();
 *				: This module can be used with a thread-chain for 
 *				: an MSL compiled to monopolize the thread.
 *	
 *	Returns a void* for use in other methods.
 */
void* RP_LFHTTP_Create(void *chain);

/*
 *	To minimize the creation of ridiculously long
 *	URI values and to help obfuscate non-shared
 *	portions of the local file system, 
 *	we must register a "starting" path.
 *	
 *	This method can be called multiple times,
 *	allowing a variety of "starting" path
 *	values to be specified.
 *
 *	rpObj				: object returned by RP_LFHTTP_Create()
 *
 *	startPath			: the starting path to register
 *
 *	Returns an integer for one of the following:
 *	zero or positive	: indicates that the 'startPath' was added
 *						: or is already registered. 
 *
 *	negative			: indicates an error
 *
 *	TODO: Customize the error code scenarios.
 */
int RP_LFHTTP_RegisterStartPath(void *rpObj, const char *startPath);

/*
 *	Sets *uri to an HTTP URI.
 *
 *	rpObj					: object returned by RP_LFHTTP_Create().
 *
 *	localFilePath			: the local file path
 *
 *	startPathIndex			: The index of the startPath that should be used.
 *							: If the app layer knows the index of the 
 *							: registered startPath (obtained from
 *							: RP_LFHTTP_RegisterStartPath), then this value can be
 *							: non-negative. Otherwise, the value can be
 *							: negative, to indicate that a full path needs
 *							: to be represented in the URI value.
 *							:
 *							: NOTE: It is always recommended that a valid
 *							: startPathIndex be provided, as the performance
 *							: penalty is huge.
 *	
 *	ipAddrAndPort			: string representation of the IP address and port
 *							: for the URI
 *
 *	ipAddrAndPortLen		: length of ipAddrAndPort
 *
 *	virDirName				: virtual directory name (e.g. base path)
 *
 *	virDirNameLen			: length of virDirName
 *
 *	uri						: if a uri can be derived and if starUriSize < 0,
 *							: then a *uri will be allocated
 *							: and a URI value will be copied to it.
 *
 *	starUriSize				: if value is negative, then *uri points to invalid memory.
 *							: if non-negative, then *uri points to memory of specified size
 *
 *	Returns the following:
 *	
 *		positive			: specified startPathIndex and localFilePath are
 *							: acceptable and a "masked" URI was set on *uri.
 *							: A masked URI is a URI that does not reveal the
 *							: full local path of the file.
 *
 *		zero				: a URI was not returned
 *
 *		negative			: The localFilePath could
 *							: not be mapped to a registered path, but
 *							: *uri was set to a URI value.
 */
int RP_LFHTTP_SetUriValue(void *rpObj, const char *localFilePath, int startPathIndex, const char *ipAddrAndPort, int ipAddrAndPortLen, const char *virDirName, int virDirNameLen, /*OUT or INOUT*/char **uri, int starUriSize);

/*
 *	Sets res->ProtocolInfo (and res->Allocated) given the local file path and file extension.
 *
 *	rpObj					: object returned by RP_LFHTTP_Create()
 *
 *	localFilePath			: the local file path
 *
 *	fileExtension			: file extension stored in localFilePath
 *
 *	res						: the CdsMediaResource object
 */
int RP_LFHTTP_SetProtocolInfo(void *rpObj, const char *localFilePath, const char *fileExtension, /*INOUT*/struct CdsMediaResource *res, int transcode_flag);

/*
 *	Method provides the appropriate local path, specified in the URI.
 *
 *	rpObj					: object returned by RP_LFHTTP_Create()
 *
 *	uriPath					: the URI path, specified in the HTTP request
 *
 *	uriPathLen				: string length of uriPath
 *
 *	requireRegisteredPath	: Nonzero value means that the URI must map to a registered path.
 *							: If the URI does not map to a registered path, then
 *							: a negative value is returned.
 *
 *	fullPath				: a block of allocated memory to store the local path name
 *
 *	fullPathSize			: number of bytes allocated for fullPath
 *
 *  transcode_flag          :0 -> no transcode, 1 -> open transcode
 *
 *	Returns positive number to indicate the fullSize needed for fullPath.
 *	If returned value is smaller than fullPathSize, then fullPath contains
 *	the local path represented by the uriPath. Returns negative number
 *	if a path could not be generated.
 *
 *	This method does not check to see if fullPath actually mapped to a local
 *	path. 
 *
 */
int RP_LFHTTP_RelativeUriToFullPath(void *rpObj, const char *uriPath, int uriPathLen, int requireRegisteredPath, /*INOUT*/ char *fullPath, int fullPathSize);

/*
 *	Use this method to respond to an HTTP request.
 *	Usually called from the dispatch sink that handles HTTP requests.
 *
 *		session							: The HTTP session for carrying the response;
 *										: obtained when your callback executes for
 *										: handling an HTTP request
 *		
 *		header							: The HTTP headers of the request.
 *										: Obtained from callback that handles the HTTP request.
 *
 *		bufferSize						: the read/transmit size for sending the file
 *
 *		filename						: the local file to send
 *
 *		contentType						: the MIME-TYPE value of the file
 *
 *		fourthField						: the 4th field of the protocolInfo value
 *
 *		onDestroyTs						: the callback to execute when the returned value is destroyed
 *
 *	Returns an object to represent the transfer. This object is to be treated
 *	as a READ-ONLY object. In fact, you should only access 
 *	ts->TotalBytesToSend and ts->TotalBytesToSend to monitor progress.
 */
struct TS_TransferState* RP_LFHTTP_Respond(struct ILibWebServer_Session *session, const struct packetheader *header, unsigned long long bufferSize, const char *filename, const char *contentType, TS_Callback_OnDestroy_TransferState onDestroyTs);

/*
 *	Protected methods.
 */
void RP_LFHTTP_GetStartPaths(void *rpObj, /*OUT*/char ***path, /*OUT*/int **pathLen, /*OUT*/int *numPaths);
const char* RP_GetProtocolInfo4thField(const char *localFilePath, const char *mimeType);
#endif

