#ifndef _TRANSFER_STATE_H
#define _TRANSFER_STATE_H

#include "WinAio.h"
#include "ILibParsers.h"
#include "ILibWebServer.h"

#ifdef _POSIX
#include <semaphore.h>
#endif

enum TransferStateFlags
{
	TS_F_MethodGet				= 0x00000001,		/* indicates GET request */
	TS_F_MethodHead				= 0x00000002,		/* indicates HEAD request */
	TS_F_ContentFeatures		= 0x00000004,		/* response should have contentFeatures.dlna.org*/
	TS_F_Range					= 0x00000008,		/* request specified the RANGE header. */
	TS_F_MultipleRange			= 0x00000010,		/* request specified multiple ranges . */

	TS_F_Aborted				= 0x00000100,		/* HTTP response was aborted */
	TS_F_DiskError				= 0x00000200,		/* disk error during response */

	TS_F_AllocTarget			= 0x00001000,	
	TS_F_AllocContentType		= 0x00002000,
	TS_F_AllocContentFeatures	= 0x00004000,
	TS_F_AllocFdes_aio_buf		= 0x00008000,
	
	TS_F_OpenFdes_aio_fildes	= 0x00010000,

	// albert add
	TS_F_InvalidFormat			= 0x04000000,
	TS_F_InternalError			= 0x08000000,
	TS_F_MethodNotImplemented	= 0x10000000,
	TS_F_InvalidRange			= 0x20000000,
	TS_F_BadRequest				= 0x40000000,
	TS_F_FileNotFound			= 0x80000000
};

#define TS_BUFFER_SIZE 4096

struct TS_Range
{
	unsigned long long StartIndex;
	unsigned long long BytesSent;
	unsigned long long BytesLeft;
};

/*
 *	Callback for when a transfer is destroyed.
 */
struct TS_TransferState;
typedef void (*TS_Callback_OnDestroy_TransferState) (struct TS_TransferState *ts);

struct TS_TransferState
{
	/* mapped by TransferStateFlags */
	unsigned int	Flags;

	/* use ILibQueue_DeQueue() to get the next TS_Range */
	void			*RangeQueue;

	/* total number of bytes sent */
	unsigned long long			TotalBytesSent;

	/* total number of bytes to send */
	unsigned long long			TotalBytesToSend;

	/* number of bytes to read/send each time */
	unsigned long long			BufferSize;

	/* target of specified method; relative path */
	char			*Target;
	int				TargetLen;

	/* mime type of the response */
	char			*ContentType;
	int				ContentTypeLen;

	/* 4th field of protocolInfo */
	char			*ContentFeatures;
	int				ContentFeaturesLen;

	/* for use with waio_ functions */
	struct aiocb	FileAiocb;

	/* reserved fields */
	struct ILibWebServer_Session	*Session;

	/* execute this callback when this object is about to be freed */
	TS_Callback_OnDestroy_TransferState Callback_OnDestroy;

	/* miscellaneous field for general use */
	void			*User;

	/* we use a randomly generated multi-part separator */
	char			*MultipartSeparator;
	int				MultipartSeparatorLen;

	/* ensures proper thread-safe deallocation */
	sem_t RefCountLock;
	int RefCount;

	//	Included for posterity when I tried to use Posix AIO
	//struct sigaction	SigAction;
};


/*
 *	Parses the RANGE HTTP header and adds the appropriate TS_Range objects to the queue.
 *
 *	rangeHeader				: the RANGE header field of an HTTP request
 *
 *	vrStart					: 0-based index value;  describes the 
 *							: starting index of the valid range
 *
 *	vrLen					: Number of bytes from vrStart that are in acceptable range. 
 *
 *	ts						: Adds TS_Range objects to ts->Queue
 *							: and will appropriately set ts->Flags and ts->TotalBytesToSend.
 */
void TS_AddRange(struct packetheader_field_node *phf, unsigned long long vrStart, unsigned long long vrLen, /*INOUT*/ struct TS_TransferState *ts);

/*
 *	Increment the transfer state's ref counter.
 */
void TS_RefInc(/*INOUT*/ struct TS_TransferState *ts);

/*
 *	Decrement the transfer state's ref counter.
 *	TransferState is destroyed when counter is zero.
 */
void TS_RefDec(/*INOUT*/ struct TS_TransferState *ts);

/*
 *	Initialize 'ts' in preparation for a response
 *	to an HTTP request.
 *
 *		session				: HTTP session that will carry the response
 *
 *		header				: HTTP headers of the request
 *
 *		bufferSize			: The read/transmit size.
 *							: If this value < 0, then ts->FileAiocb.aio_buf will be NULL,
 *							: and ts->Flags will not have the TS_F_AllocFdes_aio_buf set.
 *
 *		contentType			: MIME-TYPE value of the response
 *
 *		fourthField			: The 4th field of the protocolInfo
 *
 *		onDestroyTs			: when 'ts' is about it be destroyed, execute this callback
 *
 *		ts					: The following fields (and related length and ->Flags fields) 
 *							: on ts are initialized.
 *							:	->BufferSize
 *							:	->Callback_OnDestroy
 *							:	->ContentFeatures (MUSTDO)
 *							:	->ContentType
 *							:	->FileAiocb.aio_buf 
 *							:	->FileAiocb.aio_nbytes
 *							:	->FileAiocb.Callback 
 *							:	->FileAiocb.User
 *							:	->RangeQueue (created, not populated)
 *							:	->Session
 *							:	->Target
 *							:	->Flags (HTTP_STR_GET or HTTP_STR_HEAD or TS_F_MethodNotImplemented)
 *
 *		ignoreOrWAIO_Complete	: Valid for WIN32 only, of type WAIO_Complete.
 */
void TS_PrepareForHttp(struct ILibWebServer_Session *session, const struct packetheader *header, unsigned long long bufferSize, const char *contentType, const char *fourthField, TS_Callback_OnDestroy_TransferState onDestroyTs, /*INOUT*/struct TS_TransferState *ts, void* ignoreOrWAIO_Complete);

/*
 *	Parse the headers of an HTTP request and apply the infor to 'ts'.
 *	The following HTTP headers are supported.
 *		HTTP_STR_DLNA_GETCONTENTFEATURES 
 *		HTTP_STR_DLNA_GETCONTENTFEATURES_VAL
 *		HTTP_STR_RANGE
 *
 *	phf						: first HTTP header, obtain from packetheader->FirstField
 *
 *	vrStart					: 0-based index value;  describes the starting index of the valid range.
 *
 *	vrLen					: Number of bytes from vrStart that are in acceptable range. 
 *
 *	ts						: Applies changes to ->Flags and ->RangeQueue.
 */
void TS_ParseHttpHeaders(struct packetheader_field_node *phf, unsigned long long vrStart, unsigned long long vrLen, /*INOUT*/struct TS_TransferState *ts);

#endif

