/*
 *	MediaServerAbstraction
 *
 *	Architecturally, the module is above a set of DeviceBuilder-generated modules 
 *	(generated UPnP stack) and the module is below the actual implementation 
 *	logic of a UPnP MediaServer. This module essentially provides a level of
 *	abstraction from the UPnP communications, such that upper application layers
 *	can work with structs when processing UPnP requests. This module achieves this
 *	by implementing 3 roles.
 *
 *	Due to limitations in the underlying UPnP stack, this module is designed to
 *	be a singleton. MSL_CreateMediaServer() should only be called once for
 *	an entire process.
 *
 *	The first role is to provide all of the dispatch points for the generated 
 *	UPnP stack. Equivalently, callbacks for responding to a UPnP action/request come 
 *	to all of methods defined in this module. These methods are identified by the
 *	"UpnpContentDirectory_" and "UpnpConnectionManager_" prefixes. These methods
 *	translate a UPnP request into arguments for various callbacks.
 *
 *	The second role is to provide callbacks to upper application layers. These
 *	callbacks are identified by the "MSA_Callback_" prefix. 
 *
 *	The third role for this module is to provide the methods that allow the
 *	"logic behind a MediaServer" to respond to UPnP requests and update the
 *	MediaServer's state as reported on the UPnP network. These "public" methods
 *	are identified by the "MSA_" prefix.
 *
 *	-Each method call requires the object returned from MSA_CreateMediaServer.
 *
 *	-All strings for the interface are assumed to be UTF8 compliant. This means
 *	 that string arguments in callbacks will be sent in their UTF8-form
 *	 and that all strings sent in response to a CP's action are properly
 *	 encoded by the application layer in UTF8 form. 
 *
 *	THREADING NOTE
 *	The MSA is threadless. In practice, the MSA will execute callbacks on the
 *	thread that executes the Microstack thread-chain. 
 *	Upper layers that provide dispatch points for MSA callbacks are expected
 *	to return quickly, as not doing so will cause thread-starvation for UPnP 
 *	communications. The dispatch points for MSA callbacks should not perform
 *	I/O operations, time-consuming computations, or employ locks that are not
 *	sufficiently deterministic.
 */

#ifndef MEDIASERVERABSTRACTION_H
#define MEDIASERVERABSTRACTION_H

/************************************************************************************/
/* START SECTION - error messages */
#define MSA_ErrorMsg_ConnectionDoesNotExist "Specified connection does not exist."

/* END SECTION - error messages */
/************************************************************************************/

enum MSA_ErrorCodes
{
	MSA_Error_ConnectionDoesNotExist		= 706
};

enum MSA_Enum_SrcProtInfo
{
	MSA_SrcProtInfo_HTTPGET_STAR_STAR_STAR	= 0
};

/*
 *	Provides statistical info on the calls made on the media server.
 */
struct MSA_Stats
{
	unsigned int		Count_Browse;
	unsigned int		Count_GetSystemUpdateID;
	unsigned int		Count_GetSearchCapabilities;
	unsigned int		Count_GetSortCapabilities;
	unsigned int		Count_GetCurrentConnectionIDs;
	unsigned int		Count_GetCurrentConnectionInfo;
	unsigned int		Count_GetProtocolInfo;
};

/*
 *	Specifies the type of CDS query.
 */
enum MSA_Enum_QueryTypes
{
	MSA_Query_BrowseMetadata,
	MSA_Query_BrowseDirectChildren,
	MSA_Query_Search
};

/*
 *	Encapsulates input arguments for a Browse request.
 */
struct MSA_CdsQuery
{
	/*
	 *	Specifies what type of query.
	 */
	enum MSA_Enum_QueryTypes QueryType;

	/*
	 *	A MediaServer token. Such an object can be matched with the application's
	 *	list of active MediaServer objects, each returned from MSA_CreateMediaServer().
	 */
	void *MediaServerObject;

	/*
	 *	Application can use this token when calling UpnpResponse_xxx methods.
	 */
	void *UpnpToken;

	/*
	 *	ObjectID specified by the control point.
	 */
	char *ObjectID;

	/*
	 *	Metadata filter settings. Comma-delimited list of [namespace_prefix]:[tag_name] strings, 
	 *	that describe what CDS metadata to return in the response. In this framework,
	 *	the application layer is responsible for enforcing metadata filtering.
	 */
	char *Filter;

	/*
	 *	The index of the first media object to return. Zero-based value.
	 *	Only applicable when BrowseDirectChildren is nonzero.
	 */
	unsigned int StartingIndex;
	
	/*
	 *	The maximum number of media objects to return. Zero means return all media objects.
	 *	Only applicable when BrowseDirectChildren is nonzero.
	 */
	unsigned int RequestedCount;

	// the index that we have return back, -1 mean nothing has been returned
	unsigned int CurrentReadIndex;

	/*
	 *	SortCriteria strings have the following form:
	 *	[+/-][namespace prefix]:[tag name],[+/-][namespace prefix]:[tag name],...	 
	 */
	char *SortCriteria;

	/*
	 *	If QueryType == MS_Query_Search, then this field
	 *	specifies the search query, as specified in the format
	 *	of the CDS specification.
	 */
	char *SearchCriteria;

	/*
	 *	Miscellaneous field for application-level logic.
	 *	Application must free this field if it allocates to it.
	 */
	void *UserObject;

	/*
	 *	Provides IP address and port information 
	 *	regarding the interface that received the request.
	 */
	int RequestedOnPort;
	int RequestedOnAddress;

	/*
	 *	Provides the list of IP addresses that 
	 *	are active on this machine when the request
	 *	was received.
	 */
	int *IpAddrList;
	int IpAddrListLen;
};


/*
 *	This application-level assigned function pointer, causes
 *	the callback sink to execute when the UPnP layer 
 *	receives a Browse or Search action invocation.
 *	Applications should use the MSA_DeallocateCdsQuery()
 *	to free resources held by browseArgs.
 *
 *	msaObj				: MSA object, obtained from MSA_CreateMediaServer().
 *
 *	cdsQuery			: input args CDS metadata query;
 *						: cdsQuery->IpAddrList is ordered such that the first
 *						: IP address in the list is the one where the UPnP request
 *						: was received. This list is its own local copy.
 *						
 *	userObj				: userObj that was sent in MSA_CreateMediaServer().
 */
typedef void (*MSA_Callback_OnQuery) (void *msaObj, struct MSA_CdsQuery *cdsQuery, void *userObj);

/*
 *	This application-level assigned function pointer, causes
 *	the callback sink to execute whenever the statistics for the media server changes.
 *
 *	msaObj				: MSA object, obtained from MSA_CreateMediaServer().
 *
 *	stats				: new usage stats
 *						
 *	userObj				: userObj that was sent in MSA_CreateMediaServer().
 */
typedef void (*MSA_OnStatsChanged) (void *msaObj, struct MSA_Stats *stats, void *userObj);




/*
 *	This method creates the MediaServer object that abstracts state information
 *	for a MediaServer. The returned object will be added to the specified chain.
 *	Remember to call MSA_UpdateIpInfo() after this method.
 *
 *		chain					: A Microstack thread-chain object. see ILibCreateChain().
 *
 *		upnpStack:				: The state object for the UPnP stack. see UpnpCreateMicroStack().
 *
 *		lifetimeMonitor:		: A lifetime monitor, for timing purposes. see ILibCreateLifeTime().
 *
 *		sinkProtocolInfo:		: The comma-separated-value list of protocolInfo values, 
 *								: supported by this MediaServer as a sink. Generally an empty string ("").
 *
 *		sourceProtocolInfo		: The comma-separated-value list of protocolInfo values, 
 *								: supported by this MediaServer as a source. 
 *								: Generally the union set of protocolInfo values that
 *								: a control point will find if the entire CDS is enumerated.
 *								
 *		sortFields				: The comma-separated-value list of metadata properties
 *								: that this MediaServer can use to sort results.
 *
 *		searchFields			: The comma-separated-value list of metadata properties
 *								: that this MediaServer can use for search-based queries.
 *
 *		onQueryCallback			: MSA_Callback_OnQuery callback.
 *
 *		onStatsCallback			: MSA_OnStatsChanged callback.
 *
 *		userObj					: A miscellaneous object that is sent when callbacks are executed.
 */
void *MSA_CreateMediaServer(void *chain, void *upnpStack, void* lifetimeMonitor, const char *sinkProtocolInfo, const char *sourceProtocolInfo, const char *sortFields, const char *searchFields, MSA_Callback_OnQuery onQueryCallback, MSA_OnStatsChanged onStatsCallback, void* userObj);

/*
 *	Allows applications to deallocate the MSA_CdsQuery object
 *	that was sent in MSA_Callback_OnQuery.
 *
 *		cdsQuery				: Address of the CdsQuery specified in the callback.
 */
void MSA_DeallocateCdsQuery(struct MSA_CdsQuery *cdsQuery);

/*
 *	Application-level code can call this method within their
 *	implementation of MSA_Callback_OnQuery or MSA_Callback_OnSearch
 *	when they need to respond to the control point with an error.
 *
 *		cdsQuery				: CDS query specified in the callback.
 *
 *		errorCode				: This value can be one of the values in MSA_ErrorCodes
 *								: or it can be be a vendor-defined value between the range
 *								: of [800,899] inclusive.
 *
 *		errorMsg				: This is the custom-message to include with the error code.
 *								: The string must be encoded with UTF8 compliance.
 *
 */
void MSA_ForResponse_RespondError(struct MSA_CdsQuery *cdsQuery, int errorCode, const char *errorMsg);

/*
 *	Application-level code can call this method within their
 *	implementation of MSA_Callback_OnQuery when they need to 
 *	begin the response to the
 *	control point. This method should be called before calling 
 *	MSA_ForQueryResponse_ResultArgument or
 *	MSA_ForResponse_RespondToQuery_FinishResponse.
 *
 *		cdsQuery				: browse argument struct specified in a callback.
 *
 *		sendDidlHeader			: If nonzero, the DIDL-Lite header is sent. If nonzero,
 *								: the application is responsible for sending that header
 *								: through MSA_ForQueryResponse_ResultArgument.
 */
void MSA_ForQueryResponse_Start(struct MSA_CdsQuery *cdsQuery, int sendDidlHeader);

/*
 *	Application-level code can call this method within their
 *	implementation of MSA_Callback_OnQuery when they need to respond to the
 *	control point with data in the Result argument. This method must
 *	be called after MSA_ForResponse_RespondToQuery_StartResponse but
 *	before MSA_ForResponse_RespondToQuery_FinishResponse.
 *
 *		cdsQuery				: CDS query specified in a callback.
 *
 *		xmlEscapedUtf8Didl		: DIDL-Lite response data, where the data is properly XML-escaped and also in UTF8 encoding.
 *								: The upper application layer can call this an arbitrary number of times, so the
 *								: upper application layer can decide between making fewer calls with larger
 *								: DIDL-Lite data blocks or more calls with smaller DIDL-Lite data blocks.
 *
 *		didlSize				: The number of bytes in xmlEscapedUtf8Didl, not including a trailing null-terminator.
 */
void MSA_ForQueryResponse_ResultArgument(struct MSA_CdsQuery *cdsQuery, const char *xmlEscapedUtf8Didl, int didlSize);

/*
 *	Application-level code can call this method within their
 *	implementation of MSA_Callback_OnQuery when they need to finish the
 *	response to a control point.
 *
 *		cdsQuery				: CDS query specified in a callback.
 *
 *		sendDidlFooter			: If nonzero, the DIDL-Lite footer is sent. If nonzero,
 *								: the application will have sent the DIDL-Lite footer in a previous
 *								: call to MSA_ForQueryResponse_ResultArgument.
 *
 *		numberReturned			: the number of DIDL-Lite elements returned in the response.
 *
 *		totalMatches			: the number of DIDL-Lite elements that could be returned given the request.
 *
 *		updateID				: the updateID of the container. Zero, if respond to a BrowseMetadata request on an item.
 */
void MSA_ForQueryResponse_FinishResponse(struct MSA_CdsQuery *cdsQuery, int sendDidlFooter, unsigned int numberReturned, unsigned int totalMatches, unsigned int updateID);

/*
 *	Applications call this method to increment the mediaserver's reported
 *	value for the SystemUpdateID state variable.
 *
 *		msaObj					: The object returned from MSA_CreateMediaServer
 */
void MSA_IncrementSystemUpdateID(void *msaObj);

/*
 *	Applications call this method to instruct the mediaserver's reported
 *	UpdateID values for individual container objects. This method
 *	will automatically moderate the delivery of events for the
 *	ContainerUpdateID state variable.
 *
 *	Although this method automatically moderates the delivery of the UPnP event,
 *	applications should avoid calling this method repetitively with the same
 *	containerID in a short period of time. The primary performance issue with this
 *	method is that it has to prevent duplicate containerID entries in the
 *	state variable.
 *
 *		msaObj					: The object returned from MSA_CreateMediaServer
 *
 *		containerID				: the ID of the container that changed
 *
 *		containerUpdateID		: the new updateID of the container that changed
 */
void MSA_UpdateContainerID(void *msaObj, const char *containerID, unsigned int containerUpdateID);

/*
 *	Applications call this method to update the list of IP addresses
 *	for the mediaserver. This method needs to be called after MSA_CreateMediaServer().
 *
 *		msaObj					: The object returned from MSA_CreateMediaServer
 *
 *		ipAddrList				: Array of IP addresses
 *
 *		ipAddrListLen			: Number of entries in ipAddrList.
 *
 */
void MSA_UpdateIpInfo(void *msaObj, int *ipAddrList, int ipAddrListLen);

#endif

