#ifdef WIN32
#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#ifndef _WIN32_WCE
#include <crtdbg.h>
#endif
#else
#include <stdlib.h>
#endif

#include <stdio.h>
#include <string.h>
#include "MediaServerLogic.h"
#include "MediaServerAbstraction.h"
#include "ILibParsers.h"
#include "MyThread.h"
#include "CdsMediaObject.h"
#include "CdsObjectToDidl.h"
#include "CdsErrors.h"
#include "PortFn_MsTime.h"
#include "ILibWebServer.h"
#include "debug.h"

#ifdef OEM_ASUS
// In "DMSService.h", it is used to wait up ASUS NB
void	WaitUpSystem();
#endif

struct MSL_ResourceProvider
{
	MSL_Callback_AcquireResources OnAcquireResources;
	void* ResProviderObj;
	struct MSL_ResourceProvider *Next;
};

struct MSL_MediaServerObject
{
	/*
	 *	The PreSelect, PostSelect, and Destroy fields must
	 *	remain here as the object needs to be compatible with the
	 *	thread-chaining framework.
	 */
	void (*PreSelect)(void* object,fd_set *readset, fd_set *writeset, fd_set *errorset, int* blocktime);
	void (*PostSelect)(void* object,int slct, fd_set *readset, fd_set *writeset, fd_set *errorset);
	void (*Destroy)(void* object);

	void *MSA;
	void *UPNP;
	void *USER;

	/*
	 *	Describes the supported protocolInfo for the server, as a source.
	 *	Generally, the union set of protocolInfo values that a control
	 *	point would encounter if the entire metadata store was enumerated.
	 */
	char **SourceProtocolInfo;
	unsigned int SourceProtocolInfoCount;

	/*
	 *	Describes the supported protocolInfo for the server, as a sink.
	 *	Generally, an empty array.
	 */
	char **SinkProtocolInfo;
	unsigned int SinkProtocolInfoCount;

	/*
	 *	Describes the supported properties for sorting results
	 *	of metadata queries.
	 */
	char **SortFields;
	unsigned int SortFieldsCount;

	/*
	 *	Describes the supported properties for search-based queries.
	 */
	char **SearchFields;
	unsigned int SearchFieldsCount;

	/*
	 *	The queue for serializing CDS metadata requess.
	 */
	void *Queue;

	#ifdef MSL_THREAD_MONOPOLY
	/*
	 *	Blocks the MSA when there's nothing to do.
	 */
	sem_t LockBlock;
	int StopFlag;
	#else
	void *LTM;
	#endif

	/*
	 *	Callbacks
	 */
	MSL_Callback_OnQuery OnQuery;
	MSL_Callback_ProcessObject OnProcessObject;
	MSL_Callback_AcquireUpdateIDAndCleanup OnAcquireUpdateIDAndCleanup;
	MSL_Callback_ProcessObjects OnProcessObjects;
	MSL_Callback_OnQueryCount OnQueryCount;

	/*
	 *	Lock for the list of resource providers.
	 */
	sem_t LockResourceProviders;
	struct MSL_ResourceProvider *ResProvider_Head, *ResProvider_Tail;
};


/*
 *	Other helper functions.
 */
void _MSL_ConvertCSV(/*OUT*/unsigned int *count, /*OUT*/ char ***arrayOfValues, /*IN*/ char *csv);
void _MSL_DestroyMediaServer(void *mslObj);
void _MSL_FlushQueue(struct MSL_MediaServerObject *msl);
void _MSL_FreeCSV(/*IN*/unsigned int count, /*OUT*/ char ***arrayOfValues);
void _MSL_ProcessQueue(void* mslObj);
void _MSL_PulseProcessQueue(void* object,fd_set *socketset, fd_set *writeset, fd_set *errorset, int* blocktime);

/*
 *	Clears the queue of queries and properly deallocates them.
 */
void _MSL_FlushQueue(struct MSL_MediaServerObject *msl)
{
	struct MSA_CdsQuery *query;
	ILibQueue_Lock(msl->Queue);

	while (ILibQueue_IsEmpty(msl->Queue) == 0)
	{
		query = ILibQueue_DeQueue(msl->Queue);
		MSA_DeallocateCdsQuery(query);
	}

	ILibQueue_UnLock(msl->Queue);
}

/*
 *	Executed by the thread-chain framework, when the MSL object
 *	needs to be cleaned up.
 */
void _MSL_DestroyMediaServer(void *mslObj)
{
	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*)mslObj;
	struct MSL_ResourceProvider *resProvider = NULL, *next = NULL;

	_MSL_FreeCSV(msl->SearchFieldsCount,		&(msl->SearchFields));
	_MSL_FreeCSV(msl->SinkProtocolInfoCount,	&(msl->SinkProtocolInfo));
	_MSL_FreeCSV(msl->SortFieldsCount,			&(msl->SortFields));
	_MSL_FreeCSV(msl->SourceProtocolInfoCount,	&(msl->SourceProtocolInfo));

	/*
	 *	clear the queue of all cds queries
	 */
	_MSL_FlushQueue(msl);

	/*
	 *	destroy the queue in a thread-safe manner.
	 */
	ILibQueue_Lock(msl->Queue);
	ILibQueue_UnLock(msl->Queue);
	ILibQueue_Destroy(msl->Queue);
	msl->Queue = NULL;

	/*
	 *	destroy the resource providers
	 */
	sem_wait(&(msl->LockResourceProviders));
	resProvider = msl->ResProvider_Head;
	while (resProvider != NULL)
	{
		next = resProvider->Next;
		FREE(resProvider);
		resProvider = next;
	}
	sem_post(&(msl->LockResourceProviders));
	sem_destroy(&(msl->LockResourceProviders));

	/*
	 *	destroy the lock for blocking
	 */
	#ifdef MSL_THREAD_MONOPOLY
	sem_destroy(&(msl->LockBlock));
	#endif
}

void _MSL_FreeCSV(/*IN*/unsigned int count, /*OUT*/ char ***arrayOfValues)
{
	unsigned int i;
	for (i=0; i < count; i++)
	{
		FREE((*arrayOfValues)[i]);
		(*arrayOfValues)[i] = NULL;
	}
	free (*arrayOfValues);
	*arrayOfValues = NULL;
}

void _MSL_ConvertCSV(/*OUT*/unsigned int *count, /*OUT*/ char ***arrayOfValues, /*IN*/ char *csv)
{
	struct parser_result *parseResults;
	struct parser_result_field *prField;
	int i;

	parseResults = ILibParseStringAdv(csv, 0, (int)strlen(csv), ",", 1);
	*count = 0;
	*arrayOfValues = NULL;

	if (parseResults != NULL)
	{
		*count = parseResults->NumResults;
		if (parseResults->NumResults > 0)
		{
			*arrayOfValues = (char**) malloc (sizeof(char*) * parseResults->NumResults);
			prField = parseResults->FirstResult;
			for (i = 0; i < parseResults->NumResults; i++)
			{
				if (prField != NULL)
				{
					(*arrayOfValues)[i] = (char*) malloc (prField->datalength+1);
					memcpy((*arrayOfValues)[i], prField->data, prField->datalength);
					(*arrayOfValues)[i][prField->datalength] = '\0';
					prField = prField->NextResult;
				}
			}
		}
		ILibDestructParserResults(parseResults);
	}
}

void _MSL_PulseProcessQueue(void* object,fd_set *socketset, fd_set *writeset, fd_set *errorset, int* blocktime)
{
	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*) object;

	#ifdef MSL_THREAD_MONOPOLY
	/*
	 *	If MSL_THREAD_MONOPOLY is defined, then MSL is the only
	 *	link in the chain that uses ->PreSelect and ->PostSelect fields.
	 */
	while (msl->StopFlag == 0)
	{
		_MSL_ProcessQueue(msl);
		sem_wait(&(msl->LockBlock));
	}
	msl->StopFlag = 2;
	#else
	ILibLifeTime_AddEx(msl->LTM, msl, 500, _MSL_ProcessQueue, NULL);
	#endif

}

/*
 *	This method executes on the thread chain for the MSL.
 *	It continuously pulls CdsQueries from the queue and
 *	queues a lifetime monitor callback if there's nothing
 *	to process.
 */
void _MSL_ProcessQueue(void* mslObj)
{
	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*) mslObj;
	struct MSA_CdsQuery *cdsQuery = NULL;
	void* retQueryObj;
	struct CdsMediaObject *cdsObj = NULL;
	struct MSL_ResourceProvider *resProvider = NULL;
	char *didl;
	int didlLen;
	enum Enum_CdsErrors errorCode;
	char *errorMsg;
	unsigned int filter;
	unsigned int numberReturned = 0, totalMatched = 0, updateID = 0;

	ILibQueue_Lock(msl->Queue);
	cdsQuery = (struct MSA_CdsQuery*) ILibQueue_DeQueue(msl->Queue);
	ILibQueue_UnLock(msl->Queue);

	while ((cdsQuery != NULL) && (cdsQuery->UpnpToken != NULL))
	{
		/*
		 *	We have a metadata query, so perform the following tasks.
		 *
		 *	[1]	Execute the MSL_Callback_OnQuery callback, with the query as
		 *		an argument. The callback will return a void* for use 
		 *		in later steps.
		 *
		 *	[2] After the void* is obtained, execute the MSL_Callback_ProcessObject
		 *		callback. The callback will populate all of the fields of
		 *		the provided CdsMediaObject. The callback will also
		 *		set the ->Res field to NULL.
		 *
		 *	[3] Execute the MSL_Callback_AcquireResources callback for each of the
		 *		registered ResourceProvider objects. Each callback execution
		 *		will append zero to 'n' CdsMediaResource objects to the 
		 *		linked list from ->Res.
		 *
		 *	[4] After we have a complete CdsMediaObject with associated CdsMediaResource
		 *		objects, call CdsToDidl_GetMediaObjectDidlEscaped() to get the
		 *		XML representation of the CDS object. 
		 *
		 *	[5] Use the MSA_ForQueryResponse_ResultArgument() method to
		 *		send the XML.
		 *
		 *	[6]	Get the UpdateID value for the specified objectID.
		 */

		// albert ,(msl->OnAcquireUpdateIDAndCleanup != NULL) may be omitted
		if ((msl->OnQuery != NULL) && (msl->OnProcessObject != NULL) && (msl->OnAcquireUpdateIDAndCleanup != NULL))
		{
#ifdef OEM_ASUS
			// Call this function to wake up system if any clients try to browse the server.
			WaitUpSystem();
#endif
			filter = CdsToDidl_GetFilterBitString(cdsQuery->Filter);

			errorCode = CdsError_None;
			retQueryObj = msl->OnQuery(msl, cdsQuery, filter, msl->USER, &errorCode);
	
			if (errorCode == 0)
			{
				/*
				 *	Start success response.
				 */
				if (msl->MSA != NULL) MSA_ForQueryResponse_Start(cdsQuery, 1);

				totalMatched = msl->OnQueryCount(msl, cdsQuery, retQueryObj, filter, msl->USER);
				cdsQuery->CurrentReadIndex = cdsQuery->StartingIndex;
				if (cdsQuery->QueryType & MSA_Query_BrowseDirectChildren)
				{
					struct CdsMediaObject **cdsObjArray;
					void			**retQueryObjArray;
					unsigned int	i, count, readCount;

					// DavidHu 2006/11/29, prevent client passing wrong count 
					//if (cdsQuery->RequestedCount == 0)
					if ((cdsQuery->RequestedCount == 0) || (cdsQuery->RequestedCount > totalMatched))
					{
						// return all contents
						cdsQuery->RequestedCount = count = totalMatched;
					}
					else
					{
						count = cdsQuery->RequestedCount;
					}

					// DavidHu 2006/12/08, check whether there is requested data or not
					if (count != 0)
					{
						cdsObjArray = MALLOC(sizeof(struct CdsMediaObject *) * count);
						retQueryObjArray = MALLOC(sizeof(void *) * count);

						for (i = 0; i < count; i++)
						{
							cdsObjArray[i] = CDS_AllocateObject();
						}

						readCount = msl->OnProcessObjects(msl, cdsQuery, retQueryObjArray, filter, msl->USER, cdsObjArray, count);
					
						for (i = 0; i < readCount; i++)
						{
							/*
							*	The callback populated metadata, so ask all of the
							*	registered ResourceProviders to provide 
							*	CdsMediaResource objects.
							*/

							if ((filter & CdsFilter_ResAllAttribs) != 0)
							{
								sem_wait(&(msl->LockResourceProviders));
								resProvider = msl->ResProvider_Head;
								while (resProvider != NULL)
								{
									if (resProvider->OnAcquireResources != NULL)
									{
										resProvider->OnAcquireResources(msl, cdsQuery, retQueryObjArray[i], filter, msl->USER, resProvider->ResProviderObj, cdsObjArray[i]);
									}
									resProvider = resProvider->Next;
								}
								sem_post(&(msl->LockResourceProviders));
							}

							#ifndef METADATA_IS_XML_ESCAPED
							/*
							*	TODO: Optimize by defining METADATA_IS_XML_ESCAPED
							*	and implementing a metadata source that writes
							*	XML-escaped data.
							*/
							#define METADATA_IS_XML_ESCAPED 0
							#endif

							/*
							*	Get the XML form and send it.
							*/

							didl = CdsToDidl_GetMediaObjectDidlEscaped(cdsObjArray[i], METADATA_IS_XML_ESCAPED, filter, 0, &didlLen);
							if (didlLen > 0)
							{
								if (msl->MSA != NULL) MSA_ForQueryResponse_ResultArgument(cdsQuery, didl, didlLen);
								numberReturned++;
							}

							free (didl);
						}

						cdsQuery->CurrentReadIndex += readCount;

						for (i = 0; i < count; i++)
						{
							CDS_DestroyObjects(cdsObjArray[i]);
						}

						FREE(cdsObjArray);
						// DavidHu 2006/12/05, free the memory allocated in msl->OnProcessObjects()
						FREE(retQueryObjArray[0]);
						FREE(retQueryObjArray);
					}
					else
					{
						// There is no requested data
						errorCode = CDS_EC_OBJECT_ID_NO_EXIST;
					}
				}
				else
				{
					// MSA_Query_BrowseMetadata or MSA_Query_Search

					/* allocate an object */
					cdsObj = CDS_AllocateObject();

					if (msl->OnProcessObject(msl, cdsQuery, retQueryObj, filter, msl->USER, cdsObj))
					{
						/*
						 *	The callback populated metadata, so ask all of the
						 *	registered ResourceProviders to provide 
						 *	CdsMediaResource objects.
						 */

						if ((filter & CdsFilter_ResAllAttribs) != 0)
						{
							sem_wait(&(msl->LockResourceProviders));
							resProvider = msl->ResProvider_Head;
							while (resProvider != NULL)
							{
								if (resProvider->OnAcquireResources != NULL)
								{
									resProvider->OnAcquireResources(msl, cdsQuery, retQueryObj, filter, msl->USER, resProvider->ResProviderObj, cdsObj);
								}
								resProvider = resProvider->Next;
							}
							sem_post(&(msl->LockResourceProviders));
						}

						#ifndef METADATA_IS_XML_ESCAPED
						/*
						 *	TODO: Optimize by defining METADATA_IS_XML_ESCAPED
						 *	and implementing a metadata source that writes
						 *	XML-escaped data.
						 */
						#define METADATA_IS_XML_ESCAPED 0
						#endif

						/*
						*	Get the XML form and send it.
						*/

						didl = CdsToDidl_GetMediaObjectDidlEscaped(cdsObj, METADATA_IS_XML_ESCAPED, filter, 0, &didlLen);
						if (didlLen > 0)
						{
							if (msl->MSA != NULL) MSA_ForQueryResponse_ResultArgument(cdsQuery, didl, didlLen);
							numberReturned++;
						}
						else
						{
							/*
							 *	print an error message if we couldn't get the DIDL.
							 *	If this happens, there will likely be an inconsistency between
							 *	the number of objects returned and the TotalMatched value.
							 */
							fprintf(stderr, "_MSL_ProcessQueue() failed to serialize object %s. Reason=%d.\r\n", cdsObj->ID, didlLen);
						}

						/*
		`				 *	Free and create a new cdsobj.
						 */
						free (didl);
					}

					/* free the CDS object */
					CDS_DestroyObjects(cdsObj);
				}

				/*
				 *	Get the updateID value.
				 */
				updateID = msl->OnAcquireUpdateIDAndCleanup(msl, cdsQuery, retQueryObj, msl->USER);
				
				/*
				 *	Finish up response
				 */
				if (msl->MSA != NULL) MSA_ForQueryResponse_FinishResponse(cdsQuery, 1, numberReturned, totalMatched, updateID);
			}
			else
			{
				/*
				 *	An error has occured, so send an error response.
				 */

				switch (errorCode)
				{
				case CDS_EC_OBJECT_ID_NO_EXIST:
					errorMsg = CDS_EM_OBJECT_ID_NO_EXIST;
					break;

				case CDS_EC_NO_SUCH_CONTAINER:
					errorMsg = CDS_EM_NO_SUCH_CONTAINER;
					break;

				default:
					errorCode = CDS_EC_ACTION_FAILED;
					errorMsg = CDS_EM_ACTION_FAILED;
					break;
				}
				
				if (msl->MSA != NULL) MSA_ForResponse_RespondError(cdsQuery, errorCode, errorMsg);
			}
		}

		//pslin add to DeReference whne the session is done
		// To Release memory
		ILibWebServer_Release((struct ILibWebServer_Session *)cdsQuery->UpnpToken);

		ILibQueue_Lock(msl->Queue);
		MSA_DeallocateCdsQuery(cdsQuery);
		cdsQuery = (struct MSA_CdsQuery*) ILibQueue_DeQueue(msl->Queue);
		ILibQueue_UnLock(msl->Queue);
	}
}

void MSL_HandleMsaCallback_OnQuery (void *mslObj, struct MSA_CdsQuery *cdsQuery)
{
	/*
	 *	Take the 'cdsQuery', stick it in a queue,
	 *	and return so that the UPnP components are 
	 *	by the MSL link in the thread chain.
	 */

	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*) mslObj;
	
	//pslin add to add session reference to avoid memory fault(reserved 10, getheader)
	ILibWebServer_AddRef( (struct ILibWebServer_Session *) cdsQuery->UpnpToken );

	ILibQueue_Lock(msl->Queue);
	ILibQueue_EnQueue(msl->Queue, cdsQuery);
	ILibQueue_UnLock(msl->Queue);

	/* 
	 *	Signal the MSA thread that there's stuff to process.
	 */
	#ifdef MSL_THREAD_MONOPOLY
	sem_post(&(msl->LockBlock));
	#endif
}

void MSL_HandleMsaCallback_OnStatsChanged (void *mslObj, struct MSA_Stats *stats)
{
	/*
	 *	TODO: If your application has a GUI for reporting
	 *	usage stats, then this is the method to do that work.
	 *	Please note that this method WILL execute on the microstack
	 *	so do your work quickly and return
	 */
}

void* MSL_CreateMediaServer(void *chain, void *upnpStack, void *msaObj, char *sinkProtocolInfo, char *sourceProtocolInfo, char *sortFields, char *searchFields, 
							MSL_Callback_OnQuery queryCallback, 
							MSL_Callback_ProcessObject processObjectCallback, 
							MSL_Callback_AcquireUpdateIDAndCleanup updateIdCallback, 
							MSL_Callback_ProcessObjects processObjectsCallback, 
							MSL_Callback_OnQueryCount queryCountCallback, 
							void* userObj)
{
	struct MSL_MediaServerObject *retVal = NULL;

	/*
	 *	Allocate the MSL object
	 */
	retVal = malloc (sizeof(struct MSL_MediaServerObject));
	memset(retVal, 0, sizeof(struct MSL_MediaServerObject));

	/*
	 *	Set fields on the MSL object
	 */
	retVal->PreSelect = _MSL_PulseProcessQueue;
	retVal->Destroy = _MSL_DestroyMediaServer;
	retVal->MSA = msaObj;
	retVal->UPNP;
	retVal->USER = userObj;
	retVal->OnQuery = queryCallback;
	retVal->OnProcessObject = processObjectCallback;
	retVal->OnAcquireUpdateIDAndCleanup = updateIdCallback;
	retVal->OnProcessObjects = processObjectsCallback;
	retVal->OnQueryCount = queryCountCallback;
	sem_init(&(retVal->LockResourceProviders), 0, 1);
	retVal->Queue = ILibQueue_Create();
	#ifdef MSL_THREAD_MONOPOLY
	sem_init(&(retVal->LockBlock), 0, 1);
	retVal->StopFlag = 0;
	#else
	retVal->LTM = ILibCreateLifeTime(chain);
	#endif

	/*
	 *	Convert the CSV input params into arrays on the MSL.
	 */
	_MSL_ConvertCSV(&(retVal->SourceProtocolInfoCount), &(retVal->SourceProtocolInfo), sourceProtocolInfo);
	_MSL_ConvertCSV(&(retVal->SinkProtocolInfoCount), &(retVal->SinkProtocolInfo), sinkProtocolInfo);
	_MSL_ConvertCSV(&(retVal->SortFieldsCount), &(retVal->SortFields), sortFields);
	_MSL_ConvertCSV(&(retVal->SearchFieldsCount), &(retVal->SearchFields), searchFields);

	/*
	 *	Finished, initializing MSL object, so add it to the chain.
	 */
	ILibAddToChain(chain, retVal);


	return retVal;
}

void MSL_RegisterResourceProviderCallbacks(void *mslObj, MSL_Callback_AcquireResources onAcquireResources, void *resProviderObj)
{
	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*) mslObj;

	if (msl != NULL)
	{
		sem_wait(&(msl->LockResourceProviders));
		if (msl->ResProvider_Head == NULL)
		{
			/* empty list */
			msl->ResProvider_Tail = msl->ResProvider_Head = (struct MSL_ResourceProvider*) MALLOC(sizeof(struct MSL_ResourceProvider));
			msl->ResProvider_Tail->OnAcquireResources = onAcquireResources;
			msl->ResProvider_Tail->ResProviderObj = resProviderObj;
			msl->ResProvider_Tail->Next = NULL;
		}
		else
		{
			msl->ResProvider_Tail->Next = (struct MSL_ResourceProvider*) MALLOC(sizeof(struct MSL_ResourceProvider));
			msl->ResProvider_Tail->Next->OnAcquireResources = onAcquireResources;
			msl->ResProvider_Tail->Next->ResProviderObj = resProviderObj;
			msl->ResProvider_Tail->Next->Next = NULL;
		}
		sem_post(&(msl->LockResourceProviders));
	}
}

void MSL_ShuttingDownMicrostack(void *mslObj)
{
	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*) mslObj;
	msl->MSA = NULL;
}

void MSL_Stop(void *mslObj)
{
	#ifdef MSL_THREAD_MONOPOLY
	struct MSL_MediaServerObject *msl = (struct MSL_MediaServerObject*) mslObj;
	msl->StopFlag = 1;
	sem_post(&(msl->LockBlock));

	/* spin until it's done */
	while (msl->StopFlag != 2)
	{
		SleepMsTime(500);
	}
	#endif
}