/*
 *
 *	MediaServerLogic
 *
 *	The MediaServerLogic (MSL) provides the coordinating logic for a MediaServer
 *	implementation. Essentially, the MSL is responsible for providing the software
 *	logic that is responsible for responding to UPnP actions.
 *
 *	Architecturally, the MSL is above the MediaServerAbstraction (MSA)
 *	and below an application layer that has access to content metadata.
 *
 *	The MSL defines private-method dispatch points for MSA callbacks.
 *	The MSL also defines callbacks for the application layer to handle
 *	queries for content metadata.
 *
 *	THREADING NOTE
 *	The MSL is threadless, but it should be run on its own thread. 
 *	The reasons for this thread are described below.
 *
 *		The MSL serializes UPnP requests into a queue. Although the generated
 *		UPnP stack will perform some amount of serialization for the MSA,
 *		the underlying UPnP expects upper application layers to respond
 *		quickly to the UPnP action. Unfortunately, querying a metadata source
 *		(databases, file systems, etc.) is generally not deterministic enough.
 *		In order to prevent thread-starvation of timing-critical UPnP 
 *		communications, the MSL will use the thread for executing MSA callbacks
 *		for putting metadata requests into a queue. The MSL thread is dedicated
 *		to processing these metadata requests.
 *	
 *	Upper application layers that process MSL callbacks may choose a couple of 
 *	models for processing the callbacks.
 *
 *		[1] The thread executing the MSL callback can do the work synchronously.
 *			This means that any time spent in a lock or in a I/O operation
 *			counts against the MSL. This is probably the easiest way to do things.
 *
 *		[2] The thread executing the MSL callback does the work asynchronously.
 *			This means that the MSL callback executes and returns quickly. This
 *			model is most useful for application layers that want to employ a
 *			threadpool for processing multiple metadata requests in parallel.
 *			Generally, this model is overkill unless the expectation is for the
 *			number of metadata requests to far exceed the synchronous performance
 *			of the metadata store.
 *
 *	Lastly, the biggest reason for the MSL not encapsulating MSA is to allow
 *	the upper application layer to choose how the MSL integrates into the rest
 *	of the software solution.
 *
 *	If the upper application wants to dedicate a thread to the MSL, then
 *	the makefile should define MSL_THREAD_MONOPOLY. This will compile
 *	MSL in such a way where the MSL will have exclusive use of the thread.
 *	Obviously, the programmer must not attach any other objects to the chain
 *	that use the ->PreSelect or ->PostSelect fields.
 */


#ifndef MEDIASERVERLOGIC_H
#define MEDIASERVERLOGIC_H

#include "CdsErrors.h"
#include "CdsMediaObject.h"
#include "MediaServerAbstraction.h"

/*
 *	This callback instructs the upper application layer to perform a query.
 *
 *	mslObj				: MSL object, obtained from MSL_CreateMediaServer().
 *
 *	cdsQuery			: input args CDS metadata query;
 *						: cdsQuery->IpAddrList is ordered such that the first
 *						: IP address in the list is the one where the UPnP request
 *						: was received. This list is its own local copy.
 *
 *	filter				: bit string, mapped by 'enum CdsFilterBits'
 *						: indicating which fields should be set
 *						
 *	userObj				: userObj that was sent in MSL_CreateMediaServer().
 *
 *	cdsErrorCode		: CDS error code. see (enum Enum_CdsErrors)
 *						: If the query has ANY errors in it, a nonzero
 *						: error code value MUST be returned.
 *
 *	The callback returns a void* representing the query's results and some state
 *	information needed for the upper application layer to translate individual
 *	metadata entries into DIDL-Lite metadata. 
 */
typedef void* (*MSL_Callback_OnQuery) (void *mslObj, struct MSA_CdsQuery *cdsQuery, unsigned int filter, void *userObj, /*OUT*/ enum Enum_CdsErrors *cdsErrorCode);

/*
 *	Given a returned void* from MSL_Callback_OnQuery, this callback will execute
 *	multiple times. Each time it executes, the upper application needs to populate
 *	all fields on 'cdsObj' except cdsObj->Res. The value of cdsObj->Res must be NULL
 *	upon returning. The upper application layer can cast the 'onQueryObj' to an
 *	application-defined object, specifically designed to track the progress of 
 *	the MSL in sending metadata. In the end, MSL_Callback_ProcessObject should be 
 *	called once for every CDS object that is matched for the query.
 *
 *	mslObj				: MSL object, obtained from MSL_CreateMediaServer().
 *
 *	cdsQuery			: input args CDS metadata query;
 *						: cdsQuery->IpAddrList is ordered such that the first
 *						: IP address in the list is the one where the UPnP request
 *						: was received. This list is its own local copy.
 *
 *	onQueryObj			: object returned by the MSL_Callback_OnQuery callback.
 *
 *	filter				: bit string, mapped by 'enum CdsFilterBits'
 *						: indicating which fields should be set
 *						
 *	userObj				: userObj that was sent in MSL_CreateMediaServer().
 *
 *	cdsErrorCode		: CDS error code. see (enum Enum_CdsErrors)
 *						: If the query has ANY errors in it, a nonzero
 *						: error code value MUST be returned.
 *
 *	cdsObj				: Allocated CdsMediaObject. The dispatch point
 *						: for the callback is expected to set fields
 *						: on this CDS object. The cdsObj->Res pointer
 *						: must be NULL.
 *						: String values on CdsMediaObject need to be
 *						: UTF-8 encoded and NOT XML-escaped.
 *
 *	If metadata was populated, then the return value is nonzero.
 */
typedef int (*MSL_Callback_ProcessObject) (void *mslObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, unsigned int filter, void *userObj, /*INOUT*/ struct CdsMediaObject *cdsObj);

typedef int (*MSL_Callback_ProcessObjects) (void *mslObj, struct MSA_CdsQuery *cdsQuery, void **onQueryObj, unsigned int filter, void *userObj, /*INOUT*/ struct CdsMediaObject **cdsObj, int count);

typedef int (*MSL_Callback_OnQueryCount) (void *mslObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, unsigned int filter, void *userObj);

/*
 *	After processing the object, this callback is executed to acquire the updateID value
 *	of the object. This method is also called to deallocate the 'onQueryObj'.
 *
 *	mslObj				: MSL object, obtained from MSL_CreateMediaServer().
 *
 *	cdsQuery			: input args CDS metadata query;
 *						: cdsQuery->IpAddrList is ordered such that the first
 *						: IP address in the list is the one where the UPnP request
 *						: was received. This list is its own local copy.
 *
 *	onQueryObj			: object returned by the MSL_Callback_OnQuery callback.
 *
 *	userObj				: userObj that was sent in MSL_CreateMediaServer().
 *
 *	Return value is the 'UpdateID' value of the CDS response.
 */
typedef unsigned int (*MSL_Callback_AcquireUpdateIDAndCleanup) (void *mslObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, void *userObj);

/*
 *	After the MSL_Callback_ProcessObject is called, MSL will execute this callback
 *	multiple times, once for each registered ResourceProvider. Each ResourceProvider
 *	is expected to provide zero or more CdsMediaResource objects in the cdsObj->Res
 *	field.
 *
 *	mslObj				: MSL object, obtained from MSL_CreateMediaServer().
 *
 *	cdsQuery			: input args CDS metadata query;
 *						: cdsQuery->IpAddrList is ordered such that the first
 *						: IP address in the list is the one where the UPnP request
 *						: was received. This list is its own local copy.
 *
 *	onQueryObj			: object returned by the MSL_Callback_OnQuery callback.
 *
 *	filter				: bit string, mapped by 'enum CdsFilterBits'
 *						: indicating which fields should be set
 *
 *	resProviderObj		: 'resProviderObj' that was provided in MSL_RegisterResourceProviderCallbacks()
 *						
 *	userObj				: userObj that was sent in MSL_CreateMediaServer().
 *
 *	cdsObj				: Allocated CdsMediaObject. The dispatch point
 *						: for the callback is expected to append 
 *						: CdsMediaResource objects to the linked list
 *						: headed by cdsObj->Res.
 *						: String values on CdsMediaResource objects need to be
 *						: UTF-8 encoded, URI-escaped, and NOT XML-escaped.
 */
typedef void (*MSL_Callback_AcquireResources) (void *mslObj, struct MSA_CdsQuery *cdsQuery, void *onQueryObj, unsigned int filter, void *userObj, void *resProviderObj, /*INOUT*/ struct CdsMediaObject *cdsObj);

/*
 *	This is the handler for processing a MSA_Callback_OnQuery.
 *
 *	msaObj				: MSA object, obtained from MSA_CreateMediaServer().
 *
 *	cdsQuery			: input args CDS metadata query;
 *						: cdsQuery->IpAddrList is ordered such that the first
 *						: IP address in the list is the one where the UPnP request
 *						: was received. This list is its own local copy.
 *						
 *	userObj				: userObj that was sent in MSA_CreateMediaServer().
 */
void MSL_HandleMsaCallback_OnQuery (void *mslObj, struct MSA_CdsQuery *cdsQuery);

/*
 *	This is the handler for processing a MSA_OnStatsChanged.
 *
 *	msaObj				: MSA object, obtained from MSA_CreateMediaServer().
 *
 *	stats				: new usage stats
 *						
 *	userObj				: userObj that was sent in MSA_CreateMediaServer().
 */
void MSL_HandleMsaCallback_OnStatsChanged (void *mslObj, struct MSA_Stats *stats);

/*
 *	This method creates the MediaServer object that abstracts state information
 *	for a MediaServer. The returned object will be added to the specified chain.
 *
 *		chain					: A Microstack thread-chain object. Generally,
 *								: a different chain than the one associated
 *								: with the upnpStack. see ILibCreateChain().
 *
 *		upnpStack:				: The state object for the UPnP stack. see UpnpCreateMicroStack().
 *
 *		msaObj					: Object returned by MSA_CreateMediaServer().
 *
 *		sinkProtocolInfo:		: The comma-separated-value list of protocolInfo values, 
 *								: supported by this MediaServer as a sink. Generally an empty string ("").
 *								: Elements in the CSV must not have commas.
 *								: Furthermore, each element value must not have padded or trailing white space characters.
 *
 *		sourceProtocolInfo		: The comma-separated-value list of protocolInfo values, 
 *								: supported by this MediaServer as a source. 
 *								: Generally the union set of protocolInfo values that
 *								: a control point will find if the entire CDS is enumerated.
 *								: Elements in the CSV must not have commas.
 *								: Furthermore, each element value must not have padded or trailing white space characters.
 *								
 *		sortFields				: The comma-separated-value list of metadata properties
 *								: that this MediaServer can use to sort results.
 *								: Elements in the CSV must not have commas.
 *								: Furthermore, each element value must not have padded or trailing white space characters.
 *
 *		searchFields			: The comma-separated-value list of metadata properties
 *								: that this MediaServer can use for search-based queries.
 *								: Elements in the CSV must not have commas.
 *								: Furthermore, each element value must not have padded or trailing white space characters.
 *
 *		userObj					: A miscellaneous object that is sent when callbacks are executed.
 *
 *	If the 'chain' is different than the thread-chain associated with 'upnpStack', then
 *  the upper application layer should call ILibStartChain() once, with the 'chain'.
 *	This will allow the MSL to provide the upper application layers with a serialized
 *	set of metadata queries. ILibStartChain() will require a dedicated thread.
 */
void* MSL_CreateMediaServer(void *chain, 
							void *upnpStack, 
							void *msaObj, 
							char *sinkProtocolInfo, 
							char *sourceProtocolInfo, 
							char *sortFields, 
							char *searchFields, 
							MSL_Callback_OnQuery queryCallback, 
							MSL_Callback_ProcessObject processObjectCallback, 
							MSL_Callback_AcquireUpdateIDAndCleanup updateIdCallback,
							MSL_Callback_ProcessObjects processObjectsCallback,
							MSL_Callback_OnQueryCount queryCountCallback,
							void* userObj);

/*
 *	Tells the MSL that the Microstack is going to be shut down.
 *	The MSL will know that it's not supposed to call MSA methods.
 */
void MSL_ShuttingDownMicrostack(void *mslObj);

/*
 *	Tells the MSL to stop.
 *	Call this method before ILibStopChain().
 *
 *		mslObj					: object returned from MSL_CreateMediaServer
 */
void MSL_Stop(void *mslObj);

/*
 *	Registers ResourceProvider.
 *
 *		mslObj					: object returned from MSL_CreateMediaServer
 *
 *		onAcquireResources		: This is the dispatch point when a CdsMediaObject needs resources.
 *								: The implementation will allocate zero or more CdsMediaResource
 *								: objects and append it to ->Res field.
 *
 *		resProviderObj			: This object is provided as the 'resProviderObj' param in the callback.
 */
void MSL_RegisterResourceProviderCallbacks(void *mslObj, MSL_Callback_AcquireResources onAcquireResources, void *resProviderObj);

#endif
