#ifndef _WIN_AIO_H
#define  _WIN_AIO_H

/*
 *	WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
 *
 *	!!! DO NOT INCLUDE THIS FILE IF YOU ARE USING POSIX AIO !!!
 *
 *	WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
 */

/*
 *	This module provides a partial mapping of Posix 
 *	asynchronous I/O (AIO) interfaces for use specifically
 *	with asynchronous read operations on a local file 
 *	system.
 *
 *	POSIX NOTE:
 *	However, to get around the various nuances of
 *	using Posix signals and semaphores (see THREADING NOTE for POSIX),
 *	I've also built this API to work on Posix synchronous IO calls.
 *	This means, you no longer have a librt dependency in linux.
 *
 *	Significant differences beween WinAIO and normative Posix
 *	AIO are as follows:
 *
 *		aio_filedes is a HANDLE instead of an int
 *		aio_buf is void* instead of volatile void*
 *		aio_nbytes is DWORD instead of size_t
 *		aio_reqprio is not defined
 *		aio_sigevent is not defined
 *		aio_lio_opcode is not defined; ignored by aio_read()
 *
 *		None of the standard AIO methods are defined
 *		but some of them are wrapped, as described below.
 *
 *	THREADING NOTE:
 *	It would have been nice to implement this module without a thread,
 *	but doing so seems very risky.
 *
 *	As stated in MSDN, ReadFileEx() use a user-level asynchronous procedure call (APC)
 *	as the callback mechanism. Furthermore, the caller of ReadFileEx() also has to
 *	commit its thread for processing that APC. In order for the APC to execute,
 *	the thread needs to enter a blocking state. With this in mind, a WinAIO seems
 *	like a candidate for a link in a thread-chain.
 *
 *	Unfortunately, using the microstack's thread-chain is not an option because
 *	the APC queue processing causes undefined behavior with the select() call. 
 *	Given these types of issues, it seems best to give WinAIO its own thread
 *	for processing all of the callbacks. I may have been able to make WinAIO a
 *	link in a thread-chain but doing so may impose other restrictions on the
 *	other links in the thread-chain. 
 *
 *	THREADING NOTE for POSIX:
 *	It's clear that the problems noted in THREADING NOTE also apply to Posix.
 *	Use of the Posix AIO lib with select() yields undefined results.
 *	To make matters worse, sem_wait() is not async-signal safe so a thread can't
 *	call sem_wait() if it's going to trigger a signal. To make matters worse, the
 *	callback mechanism (sigevent.sigev_func) is not a standard posix feature, so
 *	I'm forced to use signals. This leads me to only one option that I can think of...
 *	I need to make WinAio behave like aio using synchronous IO. Yuk.
 */


#if defined(_POSIX) || defined(_WIN32_WCE)
// provides wince/posix synchronous IO support
#include "FileIoAbstraction.h"
#endif

#if defined(WIN32) && !defined(_WIN32_WCE)
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <io.h>
#include <signal.h>
#include <time.h>
#include <windows.h>
#endif

#include "MyThread.h"

/*
 *	A return value indicating that none of the requested operations 
 *	could be canceled since they are already complete. 
 */
#define AIO_ALLDONE			0x00000001

/*
 *	A return value indicating that all requested operations have been canceled. 
 */
#define AIO_CANCELED		0x00000002

/*
 *	A return value indicating that some of the requested operations could 
 *	not be canceled since they are in progress. 
 */
#define AIO_NOTCANCELED		0x00000004

/*
 *	A lio_listio() element operation option indicating that no transfer is requested. 
 */
#define LIO_NOP				0x00000008

/*
 *	A lio_listio() synchronization operation indicating that the calling 
 *	thread is to continue execution while the lio_listio() operation is 
 *	being performed, and no notification is given when the operation is complete. 
 */
#define	LIO_NOWAIT			0x00000010

/*
 *	A lio_listio() element operation option requesting a read. 
 */
#define LIO_READ			0x00000020

/*
 *	A lio_listio() synchronization operation indicating that 
 *	the calling thread is to suspend until the lio_listio() 
 *	operation is complete.
 */
#define LIO_WAIT			0x00000040
 
/*
 *	A lio_listio() element operation option requesting a write. 
 */
#define LIO_WRITE			0x00000080

#if defined(WIN32) && !defined(_WIN32_WCE)
	#define _WAIO_FDES_TYPE HANDLE
	#define _WAIO_SIZE_T unsigned long long
	#define _WAIO_ERROR DWORD

	/* these 2 originally defined when WinAio.h was supposed to be a direct API mapping of Posix AIO */
	#define _WAIO_SIGNAL_CATCHER_ARG struct aiocb*				/* type for signal catcher */
	#define _WAIO_SIGNAL_CATCHER_RET void						/* type for return value of signal catcher */
#endif

#if defined(_POSIX) || defined(_WIN32_WCE)
	#define _WAIO_FDES_TYPE void*		/* return value from FIOA_FileOpen() */
	#define _WAIO_SIZE_T unsigned long long			/* standard aio mapping */
	#define _WAIO_ERROR int				/* standard aio mapping */

	/* these 2 originally defined when WinAio.h was supposed to be a direct API mapping of Posix AIO */
	#define _WAIO_SIGNAL_CATCHER_ARG struct aiocb*				/* type for signal catcher */
	#define _WAIO_SIGNAL_CATCHER_RET void						/* type for return value of signal catcher */
#endif

struct aiocb;
typedef _WAIO_SIGNAL_CATCHER_RET (*WAIO_Complete)(_WAIO_SIGNAL_CATCHER_ARG cb);

/* minimalist implementation of Posix AIO struct sigvalue */
struct sigvalue
{
	int	sival_int;					/* ignored */
	void *sival_ptr;				/* object to send in the callback */
};

/* use with sigevent.sigev_notify - not normative to Posix. see sigevent.sigev_func. */
#define SIGEV_CALLBACK 0

/* minimalist implementation of Posix AIO struct sigevent */
struct sigevent
{
	/* set this to SIGEV_CALLBACK. If using Posix AIO signals, you would use SIGEV_SIGNAL */
	int sigev_notify;				
	int sigev_signo;				/* ignored */
	struct sigvalue sigev_value;	/* use this to set the object to send in the callback */

	/*
	 *	This is not a standard Posix thing. However, I've seen references to
	 *	this field in various man pages (such as Irix). To make things easier
	 *	to implement, I'll make use of this field. If you actually manage to
	 *	figure out how to use Posix AIO w/ signal handling, then this field is 
	 *	not needed.
	 */
	WAIO_Complete sigev_func;
};

/* minimalist implementation of Posix AIO struct sigaction : needed only if using Posix AIO signals*/
//#define SA_SIGINFO 0
//#define sigemptyset(x)
//#define sigaction(x,y,z) 
//#define SIGUSR1 0
//#define SIGEV_SIGNAL 0
//struct sigaction
//{
//	int sa_flags;					/* ignored */
//	WAIO_Complete sa_handler;		/* set the callback here */
//};

/* minimalist implementation of Posix AIO struct aiocb: used for specifying the async IO operation */
struct aiocb
{
	_WAIO_FDES_TYPE		aio_fildes;		/* File descriptor; use aio_open_RD_NONBLOCK() */
	unsigned long long				aio_offset;     /* File offset. */
	void				*aio_buf;       /* Location of buffer. */
	unsigned long long		aio_nbytes;     /* Length of transfer. */
	//int				aio_reqprio;    /* ignored; Request priority offset. */
	struct sigevent	aio_sigevent;		/* use to set the aio_sigevent.sigev_value.sival_ptr field */
	//int				aio_lio_opcode;	/* ignored: only supports READ*/
	
	/*
	 *	These fields are specific to WinAIO.
	 */
	_WAIO_SIZE_T		Reserved_dwErrorCode;
	_WAIO_SIZE_T		Reserved_Status;

	_WAIO_FDES_TYPE		aio_sync_fildes;/* file descriptor for sync read; use fopen */
	unsigned long long				aio_filesize;
    int                 isTranscoding;
};

struct WAIO_State
{
	int		StopFlag;
	sem_t	WaitForWorkerThread;
	void*	WaioWorkerThread;
	
#if defined(WIN32) && !defined(_WIN32_WCE)
	void	*OverlappedTree;
	HANDLE	TerminateEvent;
#endif

#if defined(_POSIX) || defined(_WIN32_WCE)
	void*	ApcQueue;
	sem_t	WaitForNextAPC;
	sem_t	BlockIndefinitely;
#endif
};

//	These are the functions that are part of Posix AIO that I did not implement.
//int      aio_cancel(int, struct aiocb *);		/* not supported */
//int      aio_error(const struct aiocb *);
//int      aio_fsync(int, struct aiocb *);
//int      aio_read(struct aiocb *);
//size_t  aio_return(struct aiocb *);
//int      aio_suspend(const struct aiocb *const[], int, const struct timespec *);
//int      aio_write(struct aiocb *);
//int      lio_listio(int, struct aiocb *restrict const[restrict], int,
//             struct sigevent *restrict);


#ifdef _POSIX
//
//	Included for posterity when I tried to use Posix AIO
//
// DO NOT INCLUDE <aio.h> because WinAio.h is our workaround for issues found in <aio.h>
//#include <aio.h>
//#define _WAIO_FDES_TYPE int
//#define _WAIO_SIZE_T size_t
//#define _WAIO_ERROR int
//#define _WAIO_SIGNAL_CATCHER_ARG struct aiocb*		/* type for signal catcher */
//#define _WAIO_SIGNAL_CATCHER_RET void				/* type for return value of signal catcher */
//typedef _WAIO_SIGNAL_CATCHER_RET (*WAIO_Complete)(int sig, struct siginfo* info, void * pContext);
#endif

/*
 *	Provides a wrapper for aio_error().
 */
//_WAIO_ERROR waio_error(const struct aiocb *);
#if defined(_POSIX) || defined(_WIN32_WCE)
	//Originally a direct mapping to aio_error
	//#define waio_error(cb) aio_error(cb)
	#define waio_error(cb) (cb->Reserved_dwErrorCode)
#endif
#if defined(WIN32) && !defined(_WIN32_WCE)
	#define waio_error(cb) (cb->Reserved_dwErrorCode)
#endif

#if defined(_POSIX) || defined(_WIN32_WCE)
	//Originally a direct mapping to AIO
	//#define waio_open_RD_NONBLOCK(filename) open(filename, O_RDONLY | O_NONBLOCK)
	#define waio_open_RD_NONBLOCK(filename) FIOA_FileOpen(filename, "rb")
#endif
#if defined(WIN32) && !defined(_WIN32_WCE)
	#define waio_open_RD_NONBLOCK(filename) CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL)
#endif

#if defined(_POSIX) || defined(_WIN32_WCE)
	//Originally a direct mapping to AIO
	//#define waio_close(file) close(file)
	#define waio_close(file) FIOA_FileClose(file)
#endif
#if defined(WIN32) && !defined(_WIN32_WCE)
	#define waio_close(file) CloseHandle(file)
#endif

/*
 *	Returns nonzero if the return value from waio_open is valid.
 */
#if defined(_POSIX) || defined(_WIN32_WCE)
	#define waio_fdesValid(fdes) (fdes >= 0)
#endif
#if defined(WIN32) && !defined(_WIN32_WCE)
	#define waio_fdesValid(fdes) (INVALID_HANDLE_VALUE != fdes)
#endif

#if defined(_POSIX) || defined(_WIN32_WCE)
	//Originally a direct mapping to AIO
	//#define waio_return(cb) aio_return(cb)
	#define waio_return(cb) (cb->Reserved_Status)
#endif
#if defined(WIN32) && !defined(_WIN32_WCE)
	#define waio_return(cb) (cb->Reserved_Status)
#endif

/*
 *	Provides a wrapped call to aio_read().
 */
int waio_read(struct aiocb *cb);

/*
 *	Initializes AIO state info.
 */
void waio_start();

/*
 *	Cleans up AIO state info. 
 */
void waio_stop();



#endif
