#include <stdafx.h>
#include <stdio.h>
#include <io.h>
#include <stdlib.h>
#include <direct.h>


#include "bk.hpp"
#include "bkError.hpp"
#include "bkExtract.hpp"
#include "bkInternal.hpp"
#include "bkPath.hpp"
#include "bkIoWrappers.hpp"
int bk_extract(VolInfo* volInfo, const char* srcPathAndName, 
               const char* destDir, int keepPermissions, 
               void(*progressFunction)(VolInfo*))
{
    return bk_extract_as(volInfo, srcPathAndName, destDir, NULL, 
                         keepPermissions, progressFunction);
}

int bk_extract_as(VolInfo* volInfo, const char* srcPathAndName, 
                  const char* destDir, const char* nameToUse,
                  int keepPermissions, void(*progressFunction)(VolInfo*))
{
    int rc;
    NewPath srcPath;
    BkDir* parentDir;
    int dirFound;
    
    volInfo->progressFunction = progressFunction;
    volInfo->stopOperation = 0;
    
    rc = makeNewPathFromString(srcPathAndName, &srcPath);
    if(rc <= 0)
    {
        freePathContents(&srcPath);
        return rc;
    }
    
    if(srcPath.numChildren == 0)
    {
        freePathContents(&srcPath);
        return BKERROR_EXTRACT_ROOT;
    }
    
    srcPath.numChildren--;
    dirFound = findDirByNewPath(&srcPath, &(volInfo->dirTree), &parentDir);
    srcPath.numChildren++;
    if(!dirFound)
    {
        freePathContents(&srcPath);
        return BKERROR_DIR_NOT_FOUND_ON_IMAGE;
    }
    
    rc = extract(volInfo, parentDir, srcPath.children[srcPath.numChildren - 1], 
                 destDir, nameToUse, keepPermissions);
    
    freePathContents(&srcPath);
    
    if(rc <= 0)
    {
        return rc;
    }
    
    return 1;
}

int extract(VolInfo* volInfo, BkDir* parentDir, char* nameToExtract, 
            const char* destDir, const char* nameToUse, int keepPermissions)
{
    BkFileBase* child;
    int rc;
    
    child = parentDir->children;
    while(child != NULL)
    {
        if(volInfo->stopOperation)
            return BKERROR_OPER_CANCELED_BY_USER;
        
        if(strcmp(child->name, nameToExtract) == 0)
        {
            if( IS_DIR(child->posixFileMode) )
            {
                rc = extractDir(volInfo, BK_DIR_PTR(child), destDir, nameToUse, keepPermissions);
            }
            else if ( IS_REG_FILE(child->posixFileMode) )
            {
				rc = extractFile(volInfo, BK_FILE_PTR(child), destDir, nameToUse, keepPermissions);
            }
            else if ( IS_SYMLINK(child->posixFileMode) )
            {
				//rc = extractSymlink(BK_SYMLINK_PTR(child), destDir, nameToUse);
            }
            else
            {
                rc = 1;
                printf("trying to extract something that's not a file, "
                       "symlink or directory, ignored\n");fflush(NULL);
            }
            
            if(rc <= 0)
            {
                int goOn;
                
                if(volInfo->warningCbk != NULL && !volInfo->stopOperation)
                /* perhaps the user wants to ignore this failure */
                {
                    sprintf(volInfo->warningMessage,  
                             "Failed to extract item '%s': '%s'",
                             child->name, 
                             bk_get_error_string(rc));
                    
					goOn = volInfo->warningCbk(volInfo->warningMessage);
                    rc = BKWARNING_OPER_PARTLY_FAILED;
                }
                else
                    goOn = 0;
                
                if(!goOn)
                {
                    volInfo->stopOperation = 1;
                    return rc;
                }
            }
        }
        
        child = child->next;
    }
    
    return 1;
}


int extractDir(VolInfo* volInfo, BkDir* srcDir, const char* destDir, 
               const char* nameToUse, int keepPermissions)
{
    int rc;
    BkFileBase* child;
    
    /* vars to create destination dir */
    char* newDestDir;
    unsigned destDirPerms;
    
    /* CREATE destination dir on filesystem */
    /* 1 for '\0' */
    if(nameToUse == NULL)
        newDestDir = (char*)malloc(strlen(destDir) + strlen(BK_BASE_PTR(srcDir)->name) + 2);
    else
        newDestDir = (char *)malloc(strlen(destDir) + strlen(nameToUse) + 2);
    if(newDestDir == NULL)
        return BKERROR_OUT_OF_MEMORY;
    
    strcpy(newDestDir, destDir);
    if(destDir[strlen(destDir) - 1] != '/')
        strcat(newDestDir, "/");
    
    if(nameToUse == NULL)
        strcat(newDestDir, BK_BASE_PTR(srcDir)->name);
    else
        strcat(newDestDir, nameToUse);
    
    if(keepPermissions)
        destDirPerms = BK_BASE_PTR(BK_BASE_PTR(srcDir))->posixFileMode;
    else
        destDirPerms = volInfo->posixDirDefaults;
    /* want to make sure user has write and execute permissions to new directory 
    * so that can extract stuff into it */
    destDirPerms |= 0300;
    
    if(access(newDestDir, F_OK) == 0)
    {
        free(newDestDir);
        return BKERROR_DUPLICATE_EXTRACT;
    }
    
    rc = mkdir(newDestDir);//, destDirPerms);
	rc = chmod(newDestDir,destDirPerms);
    if(rc == -1)
    {
        free(newDestDir);
        return BKERROR_MKDIR_FAILED;
    }
    /* END CREATE destination dir on filesystem */
    
    /* EXTRACT children */
    child = srcDir->children;
    while(child != NULL)
    {
        rc = extract(volInfo, srcDir, child->name, newDestDir, 
                     NULL, keepPermissions);
        if(rc <= 0)
        {
            free(newDestDir);
            return rc;
        }
        
        child = child->next;
    }
    /* END EXTRACT children */
    
    free(newDestDir);
    
    return 1;
}

int extractFile(VolInfo* volInfo, BkFile* srcFileInTree, const char* destDir, 
                const char* nameToUse, int keepPermissions)
{
    HANDLE srcFile;
    int srcFileWasOpened;
    char* destPathAndName;
    unsigned  destFilePerms;
    HANDLE destFile; /* returned by open() */
    int rc;
    BkStatStruct statStruct;

    if(srcFileInTree->onImage)
    {
        srcFile = volInfo->imageForReading;
        bkSeekSet(volInfo->imageForReading, srcFileInTree->position, SEEK_SET);
        srcFileWasOpened = 0;
    }
    else
    {
		srcFile = CreateFile(srcFileInTree->pathAndName,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
//		srcFile=CreateFile(
//        srcFile = open(srcFileInTree->pathAndName, _O_RDONLY, 0);
        if(srcFile == NULL)
            return BKERROR_OPEN_READ_FAILED;
        srcFileWasOpened = 1;
        
        /* UPDATE the file's size, in case it's changed since we added it */
        rc = bkStat(srcFileInTree->pathAndName, &statStruct);
        if(rc != 0)
            return BKERROR_STAT_FAILED;
        
        if(statStruct.st_size > 0xFFFFFFFF)
        /* size won't fit in a 32bit variable on the iso */
            return BKERROR_EDITED_EXTRACT_TOO_BIG;
        
        srcFileInTree->size = statStruct.st_size;
        /* UPDATE the file's size, in case it's changed since we added it */
    }
    
    if(nameToUse == NULL)
        destPathAndName = (char*)malloc(strlen(destDir) + 
                                 strlen(BK_BASE_PTR(srcFileInTree)->name) + 2);
    else
        destPathAndName = (char *)malloc(strlen(destDir) + strlen(nameToUse) + 2);
    if(destPathAndName == NULL)
    {
        if(srcFileWasOpened)
            bkClose(srcFile);
        return BKERROR_OUT_OF_MEMORY;
    }
    
    strcpy(destPathAndName, destDir);
    if(destDir[strlen(destDir) - 1] != '/')
        strcat(destPathAndName, "//");
    if(nameToUse == NULL)
        strcat(destPathAndName, BK_BASE_PTR(srcFileInTree)->name);
    else
        strcat(destPathAndName, nameToUse);
    /*
    if(access(destPathAndName, F_OK) == 0)
    {
        if(srcFileWasOpened)
            bkClose((int)srcFile);
        free(destPathAndName);
        return BKERROR_DUPLICATE_EXTRACT;
    }
   */ 
    
	
    if(keepPermissions)
        destFilePerms = BK_BASE_PTR(srcFileInTree)->posixFileMode;
    else
        destFilePerms = volInfo->posixFileDefaults;

	destFile = 	CreateFile(destPathAndName,GENERIC_WRITE,FILE_SHARE_WRITE,NULL,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,NULL);

 //   destFile = fopen(,"wb");
		//open(destPathAndName, O_WRONLY | O_CREAT | O_TRUNC, destFilePerms);
    if(destFile == NULL)
    {
        if(srcFileWasOpened)
            bkClose(srcFile);
        free(destPathAndName);
        return BKERROR_OPEN_WRITE_FAILED;
    }
    
  //  free(destPathAndName);
    
    rc = copyByteBlock(volInfo, srcFile, destFile, srcFileInTree->size);
    if(rc < 0)
    {
        bkClose(destFile);
        if(srcFileWasOpened)
            bkClose(srcFile);
        return rc;
    }
    
    bkClose(destFile);
    if(destFile == NULL)
    {
        if(srcFileWasOpened)
            bkClose(srcFile);
        return BKERROR_EXOTIC;
    }
    
    
    if(srcFileWasOpened)
        bkClose(srcFile);
    
    return 1;
}


int copyByteBlock(VolInfo* volInfo, HANDLE src, HANDLE dest, unsigned numBytes)
{
    int rc;
    int count;
    int numBlocks;
    int sizeLastBlock;
    
    numBlocks = numBytes / READ_WRITE_BUFFER_SIZE;
    sizeLastBlock = numBytes % READ_WRITE_BUFFER_SIZE;
    
//    maybeUpdateProgress(volInfo);
    if(volInfo->stopOperation)
        return BKERROR_OPER_CANCELED_BY_USER;
    LPBYTE ptr;
    for(count = 0; count < numBlocks; count++)
    {
//        maybeUpdateProgress(volInfo);
        if(volInfo->stopOperation)
            return BKERROR_OPER_CANCELED_BY_USER;
        ptr=(LPBYTE)volInfo->readWriteBuffer;
        rc = bkRead(src, ptr, READ_WRITE_BUFFER_SIZE);
        if(rc != READ_WRITE_BUFFER_SIZE)
            return BKERROR_READ_GENERIC;
		
        rc = bkWrite(dest, ptr, READ_WRITE_BUFFER_SIZE);
        if(rc <= 0)
            return rc;
    }
    
    if(sizeLastBlock > 0)
    {
		ptr= (LPBYTE) volInfo->readWriteBuffer;
		rc = bkRead(src, ptr, sizeLastBlock);
        if(rc != sizeLastBlock)
            return BKERROR_READ_GENERIC;
        
		rc = bkWrite(dest, ptr, sizeLastBlock);
        if(rc <= 0)
            return rc;
    }
    
    return 1;
}