/***********************************************************************
 *    rtsp.c: for downloading via rtsp (Real Time Streaming Protocol)
 ***********************************************************************
 * Copyright (C) 2007 metro <me_t_ro@yahoo.com>
 *
 * This file is part of msdl, media stream downloader
 *
 * This file is based on rtsp implementation of 
 * mplayer and xine, and includes code from both projects.
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *
 ***********************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include <unistd.h>

//#include <inttypes.h>
#include <sys/types.h>
#include<winsock2.h>

#include "msdl.h"
#include "msdllib.h"
#include "network.h"
#include "rtsp.h"
#include "real.h"
#include "rmff.h"
#include "asf.h"
#include "wmserver.h"
//#include <sys/syscall.h>

//#define strncasecmp strnicmp


const char rtsp_protocol_version[] = "RTSP/1.0";


/* realplayer specific stuff */
const char real_clientid[] = "ClientID: Linux_2.6_10.0.0.0_play32_RN01_EN_586";
const char real_useragent[] = "User-Agent: RealMedia Player Version 10.0.0.0 (linux-2.6-libc6-gcc4.1-i586)";
const char real_companyid[] = "CompanyID: X160hPo6JmsD6Vger24oAg==";
const char real_clientchallenge[] = "ClientChallenge: cd7cb1ac431e8d1ad4d2fadb8cf762d1";
const char real_playerstarttime[] = "PlayerStarttime: [19/03/2007:21:42:56 00:00]";
const char real_transport[] = "Transport: x-pn-tng/tcp;mode=play,rtp/avp/tcp;unicast;mode=play";

static struct rtsp_ctrl_t *new_rtsp_ctrl_t(void);
static void free_rtsp_ctrl_t(struct rtsp_ctrl_t *ctrlt);

int rtsp_parse_response(struct rtsp_ctrl_t *rtsp_ctrl,
			struct rtsp_header_t *rtsp_header);

char *rtsp_request_options(struct rtsp_header_t *rtsp_hdr,struct URL_t *,
			   char *what);




/*
 * read rtsp stream. filles buffer, and buffer size is 'size' 
 *
 *        read cached data from stream->stream_ctrl->buffer
 *        check for to see if network access is necessary
 *        get chunk(media packet) from network.
 *
 *  return value: bytes written to buffer. -1 if error.
 */
int rtsp_streaming_read(struct stream_t *stream,
			unsigned char *buffer, size_t buffer_size)
{
  
    struct stream_ctrl_t *stream_ctrl = stream->stream_ctrl;
    size_t pos = 0; // how many bytes are in 'buffer'
    int ret = 0;

    if(buffer_size == 0) {
	display(MSDL_ERR,"buffer_size must be bigger than 0\n");
	return -1;
    }
  
    if(stream_ctrl->write_data_len) { // there are waiting data to be written
    
	if(buffer_size <= stream_ctrl->write_data_len) {
	    /*
	      buffer_size < stream_ctrl->write_data_len...
	      fill buffer, and return.
	    */
	    memcpy(buffer,stream_ctrl->write_buffer + stream_ctrl->write_pos,
		   buffer_size);
	    stream_ctrl->write_data_len -= buffer_size;
	    stream_ctrl->write_pos += buffer_size;
	    return buffer_size;
	}
	else { 
	    /*
	      stream_ctrl->write_data_len < buffer_size,
	      have to read from network.
	    */
	    memcpy(buffer,stream_ctrl->write_buffer + stream_ctrl->write_pos,
		   stream_ctrl->write_data_len);
	    pos = stream_ctrl->write_data_len;
	    /* and stream_ctrl->write_buffer is empty. */
	    stream_ctrl->write_data_len = 0;
	    stream_ctrl->write_pos = 0;
	}
    }
  

    /* still have to get data from network. */
    if(stream_ctrl->status != streaming_finished) { /* not finished */
	//int i;
	if(stream_ctrl->rtsp_ctrl->get_media_packet == NULL) {
	    /* default use real function */
	    ret = real_rdt_get_media_packet(stream,buffer + pos,
					    buffer_size - pos);
	}
	else {
	    ret = stream_ctrl->rtsp_ctrl->get_media_packet(stream,buffer + pos,
							   buffer_size - pos);
	}
    }
  
    if(ret >= 0) return (ret + pos);
    else         return -1; /* error */
  
}







/*
 * receive rtsp message from stream->sock.
 *      header is guraranteed to be complete after this function,
 *      body should not be complete.
 *      so search Content-length and get all the body later.
 *               return status code : success
 *                               -1 : failure
 */
int rtsp_recv_header(struct stream_t *stream, struct rtsp_header_t *rtsp_hdr)
{
    struct stream_ctrl_t *stream_ctrl = stream->stream_ctrl;
    int ret,i,total;
  
    rtsp_hdr->buffer_len = 0;
    rtsp_hdr->buffer = NULL;
    total = 0;
  
    do { // get rtsp reply
	rtsp_hdr->buffer_len += BUFSIZE_1K;
	rtsp_hdr->buffer = (char *)xrealloc(rtsp_hdr->buffer,
					    rtsp_hdr->buffer_len + 1);

	i = recv_data(stream,rtsp_hdr->buffer + total,BUFSIZE_1K);

	if(i <= 0) {
	    display(MSDL_ERR,"recv error: recv_data() returned %d\n",i);
	    goto failed;
	}
	total += i;
    
	rtsp_hdr->buffer[total] = '\0'; /* for rtsp_is_entire_header */
    } while(!rtsp_is_entire_header(rtsp_hdr));

  
    /* rtsp_hdr->buffer_size is length in buffer, not the malloc()ed size.  */
    rtsp_hdr->buffer_len = total;
    rtsp_hdr->buffer[total] = '\0';
  
    ret = rtsp_parse_response(stream_ctrl->rtsp_ctrl,rtsp_hdr);
  
    if(ret < 0) {
	display(MSDL_ERR,"response RTSP header parse failed\n");
	goto failed;
    }

    /* push body back !!! */
    if(rtsp_hdr->body_len) {
	stream_data_push_back(stream,rtsp_hdr->body,rtsp_hdr->body_len);
    }
  
    /* rtsp_hdr->buffer *ONLY* contains header, no body follows */
    memset(rtsp_hdr->body,0,rtsp_hdr->body_len);

    display(MSDL_DBG,"rtsp header===========\n"
	    "%s\n"
	    "==(%d bytes)====================\n",
	    rtsp_hdr->buffer,strlen(rtsp_hdr->buffer));
  
    /* success */
    return rtsp_hdr->status_code;
  
  failed:
    if(rtsp_hdr->buffer) free(rtsp_hdr->buffer);
    rtsp_hdr->buffer = NULL;
    return -1;
}



/*
 * set fields which are common in all rtsp packet. 
 * CSeq and Session field should be present in
 * every rtsp packet.
 */
int rtsp_set_standard_fields(struct rtsp_ctrl_t *rtsp_ctrl,
			     struct rtsp_header_t *rtsp_hdr)
{
    char buf[20];
    _snprintf(buf,20,"Cseq: %u",rtsp_ctrl->cseq);
    rtsp_set_field(rtsp_hdr,buf);
  
    if(rtsp_ctrl->session) {
	char *buf2;
	buf2 = (char *)xmalloc(strlen(rtsp_ctrl->session) + 15);
	sprintf(buf2,"Session: %s",rtsp_ctrl->session);
	rtsp_set_field(rtsp_hdr,buf2);
	free(buf2);
    }

    return 0;
}


/*
 * send rtsp header (rtsp_hdr->buffer)
 */
int rtsp_send_request(struct stream_t *stream,
		      struct rtsp_header_t *rtsp_hdr)
{
    int ret;
  
    /* dbg */
    display(MSDL_DBG,"SEND rtsp header ------------>\n"
	    "%s\n--(%d bytes)---------------<\n",
	    rtsp_hdr->buffer,(int)rtsp_hdr->buffer_len);
    
    ret = xsend(stream->sock,rtsp_hdr->buffer,rtsp_hdr->buffer_len);
  
    /* OK change stream->stream_ctrl->rtsp_ctrl->cseq!! */
    stream->stream_ctrl->rtsp_ctrl->cseq++;
  
    return ret;
}



/*
 * set basic stuff and send OPTIONS request.
 * this can be used to determine  which protocol to use,
 * depending on its server.
 *              return value    rtsp_hdr : success
 *                                  NULL : failure
 */
struct rtsp_header_t *rtsp_connect(struct stream_t *stream)
{
    int sock;
    int ret;
    struct stream_ctrl_t *stream_ctrl = stream->stream_ctrl;
    struct rtsp_ctrl_t *rtsp_ctrl = stream_ctrl->rtsp_ctrl;
    struct URL_t *url = stream->stream_ctrl->url;

    struct rtsp_header_t *rtsp_hdr = NULL;

    
    if(url->port == 0) {
	url->port = RTSP_PORT; /* RTSP default port */
    }
  
  
    rtsp_ctrl->server = NULL;
    rtsp_ctrl->cseq = 1;
    rtsp_ctrl->server_state = 0;
    rtsp_ctrl->server_caps  = 0;
    rtsp_ctrl->session = NULL;
    rtsp_ctrl->get_media_packet = NULL;


    sock = server_connect(url->hostname,url->port);
    if(sock < 0) { /* couldn't connect for some reason. (port closed maybe) */
	goto failed;
    }
    stream->sock = sock; /* save socket id */

    /* default is rtsp-real (becasue OPTIONS req is supported) */
    rtsp_hdr = new_rtsp_header();
    rtsp_set_standard_fields(rtsp_ctrl,rtsp_hdr);
  
    rtsp_set_field(rtsp_hdr,real_useragent);
    rtsp_set_field(rtsp_hdr,"GUID: 00000000-0000-0000-0000-000000000000");
    rtsp_set_field(rtsp_hdr,real_clientid);  
    rtsp_set_field(rtsp_hdr,"Pragma: initiate-session");
    rtsp_set_field(rtsp_hdr,"RegionData: 0");
    rtsp_set_field(rtsp_hdr,real_clientchallenge);
    rtsp_set_field(rtsp_hdr,real_companyid);
    rtsp_set_field(rtsp_hdr,real_playerstarttime);
  
    rtsp_request_options(rtsp_hdr,url,NULL);
    rtsp_send_request(stream,rtsp_hdr);
    free_rtsp_header(rtsp_hdr);
  
    rtsp_hdr = new_rtsp_header();
    ret = rtsp_recv_header(stream,rtsp_hdr);
    if(ret < 0) {
	goto failed;
    }
  
    /* make mrl */
    rtsp_make_mrl(rtsp_ctrl,url);
  
    return rtsp_hdr;
  
  failed:
    if(rtsp_hdr) free_rtsp_header(rtsp_hdr);
    return NULL;
}




/*
 * make mrl from url and set to rtsp_ctrl->mrl 
 * mrl is "rtsp://server.com:port/path/to/file.foo"
 */
int rtsp_make_mrl(struct rtsp_ctrl_t *rtsp_ctrl,struct URL_t *url)
{
    int len;
    len = strlen(url->hostname) + strlen(url->filepath) + 32; /* 32 is enough */
  
    rtsp_ctrl->mrl = (char *)xmalloc(len);
    _snprintf(rtsp_ctrl->mrl,len,"rtsp://%s:%i%s",
	     url->hostname,url->port,url->filepath);
    return 0;
}



int rtsp_parse_response(struct rtsp_ctrl_t *rtsp_ctrl,
			struct rtsp_header_t *rtsp_hdr)
{
    char *field;
    int answer_cseq;

    if(rtsp_response_parse(rtsp_hdr) < 0) {
	display(MSDL_ERR,"RTSP header parse error\n");
	return -1;
    }
  
    if((field = rtsp_get_field(rtsp_hdr,"Alert")) != NULL) {
	display(MSDL_ERR,"*** Alert ***\n%s\n",field);
    }

  
    if((field = rtsp_get_field(rtsp_hdr,"Cseq")) != NULL) {
	sscanf(field,"%d",&answer_cseq);
	if(rtsp_ctrl->cseq - 1 != answer_cseq) {
	    /* -1 because cseq is incremented in build_request */
	    display(MSDL_DBG | MSDL_ERR,"warning: Cseq mismatch, got %u, expected %u\n",
		    answer_cseq,rtsp_ctrl->cseq - 1);
	}
    }
  
    if((field = rtsp_get_field(rtsp_hdr,"Server")) != NULL) {
	while(*field == ' ') field++;
	if(rtsp_ctrl->server) { /* already has it */
	    if(strcmp(rtsp_ctrl->server,field)) { /* different server */
		display(MSDL_VER | MSDL_ERR,"warning: got different server\n"
			"    old: %s\n"
			"    new: %s\n",rtsp_ctrl->server,field);
		free(rtsp_ctrl->server);
		rtsp_ctrl->server = strdup(field);
	    }
	}
	else { /* not set yet */
	    rtsp_ctrl->server = strdup(field);
	}
    }
  
    if((field = rtsp_get_field(rtsp_hdr,"Session")) != NULL) {
	while(*field == ' ') field++;
	if(rtsp_ctrl->session) {
	    if(strcmp(rtsp_ctrl->session,field)) { /* different session */
		display(MSDL_VER | MSDL_ERR,"warning: got different session\n"
			"    old: [%s]\n"
			"    new: [%s]\n",rtsp_ctrl->session,field);
		free(rtsp_ctrl->session);
		rtsp_ctrl->session = strdup(field);
	    }
	}
	else {
	    rtsp_ctrl->session = strdup(field);
	}
    }
  
    /* real only, ignore in other protocol... */
    if((field = rtsp_get_field(rtsp_hdr,"RealChallenge1")) != NULL) {
	while(*field == ' ') field++;
	if(rtsp_ctrl->challenge) {
	    if(strcmp(rtsp_ctrl->challenge,field)) { /* different server */
		display(MSDL_ERR,"warning: got different client challenge\n"
			"    old: %s\n"
			"    new: %s\n",rtsp_ctrl->challenge,field);
		free(rtsp_ctrl->challenge);
		rtsp_ctrl->challenge = strdup(field);
	    }
	}
	else {
	    rtsp_ctrl->challenge = strdup(field);
	}
    }
    
    return 0;
}



/*
 * rtsp_response_parse :  parse rtsp response header 
 * make rtsp_hdr from raw data in rtsp_hdr->buffer
 *
 *    return value :        -1 ... ERROR
 *                           0 ... already parsed.
 *                           1 ... success.
 */
int rtsp_response_parse(struct rtsp_header_t *rtsp_hdr)
{
  
    char *hdr_ptr, *ptr;
    char *field = NULL;
  
    int pos_hdr_sep, hdr_sep_len;
    size_t len;
  
    if(rtsp_hdr == NULL) return -1;
    if(rtsp_hdr->is_parsed) return 0; // already parsed.
  
    /* get first word  */
    hdr_ptr = strstr(rtsp_hdr->buffer, " ");
    if(hdr_ptr == NULL) {
	display(MSDL_ERR,"Malformed answer : No %20 separator\n");
	return -1;
    }
  
    len = hdr_ptr - rtsp_hdr->buffer;
    rtsp_hdr->protocol = xmalloc(len + 1);
    strncpy(rtsp_hdr->protocol,rtsp_hdr->buffer,len);
    rtsp_hdr->protocol[len] = '\0';
  

    /* 1. set method */
    rtsp_hdr->method = strdup(rtsp_hdr->protocol);
  
    if(!strnicmp(rtsp_hdr->protocol,"RTSP",4)) { /* RTSP message*/

    
	/* 2. get RTSP status code */
	if(sscanf(++hdr_ptr,"%d",&(rtsp_hdr->status_code)) != 1) {
	    display(MSDL_ERR,"Malformed answer : No rtsp status code!!\n");
	    return -1;
	} 
    
	hdr_ptr += 4; /* "[0-9][0-9][0-9] " */
    
	/* 3. get reason phrase */
	ptr = strstr(hdr_ptr,"\n");
	if(ptr == NULL) {
	    display(MSDL_ERR,"Malformed answer : unable to get reason_phrase\n");
	    return -1;
	}
	len = ptr - hdr_ptr; // len of reason phrase.

	rtsp_hdr->reason_phrase = xmalloc(len + 1);
	strncpy(rtsp_hdr->reason_phrase,hdr_ptr,len);
	if(rtsp_hdr->reason_phrase[len - 1] == '\r') { // M$ style newline!
	    len --; // we don't need \r.
	}
	rtsp_hdr->reason_phrase[len] = '\0'; // terminate string
    } 
    else if(!strnicmp(rtsp_hdr->protocol,"SET_PARAMETER",strlen("SET_PARAMETER"))) {
	rtsp_hdr->status_code = RTSP_STATUS_SET_PARAMETER;
    }
    else if(!strnicmp(rtsp_hdr->protocol,"OPTIONS",strlen("OPTIONS"))) {
	
	rtsp_hdr->status_code = RTSP_STATUS_OPTIONS;
    }
    else if(!strnicmp(rtsp_hdr->protocol,"ANNOUNCE",strlen("ANNOUNCE"))) {

	rtsp_hdr->status_code = RTSP_STATUS_ANNOUNCE;
    }
    else {  /* NOT RTSP ---> abort */
	display(MSDL_ERR,"Protocol not RTSP, server says '%s'",rtsp_hdr->protocol);
	return -1;
    }
  
  
    hdr_sep_len = 4; // header separator length
    ptr = strstr(rtsp_hdr->buffer,"\r\n\r\n");
    if(!ptr) {
	ptr = strstr(rtsp_hdr->buffer,"\n\n");
	if(!ptr) {
	    display(MSDL_ERR,"Header may be incomplete!\n");
	    return -1;
	}
	hdr_sep_len = 2;
    }
    pos_hdr_sep = ptr - rtsp_hdr->buffer;
  
    /* points to first line after method line. */
    hdr_ptr = strstr(rtsp_hdr->buffer,"\n") + 1;

    /* parse of first line [DONE] */
  
  
    /* get all fields. they are separated by '\n' */
    do {
	ptr = hdr_ptr;
	while(*ptr != '\r' && *ptr != '\n') ptr++;
	len  = ptr - hdr_ptr;

	if(len == 0) break; /* met \n\n --> end of header! */
    
	field = (char *)xrealloc(field,len + 1); // +1 for '\0'
	strncpy(field,hdr_ptr,len);
	field[len] = '\0';
    
    
	rtsp_set_field(rtsp_hdr,field);
	hdr_ptr = ptr + ((*ptr == '\r') ? 2 : 1); /* points to next line */
    } while(hdr_ptr < (rtsp_hdr->buffer + pos_hdr_sep));
  
    if(field) {
	free(field);
    }

    /*
      rtsp_hdr->buffer_len is length of header total data received
    */
    if(pos_hdr_sep + hdr_sep_len < rtsp_hdr->buffer_len) {
	rtsp_hdr->body = rtsp_hdr->buffer + (pos_hdr_sep + hdr_sep_len);
	rtsp_hdr->body_len = rtsp_hdr->buffer_len - (pos_hdr_sep + hdr_sep_len);
    }
  
    /* mark parsed */
    rtsp_hdr->is_parsed = 1;
    
  
    return 1;
}





/*
 * start rtsp streaming.
 * 
 *    return value :   negative or 0  ... error
 *                                 1  ... success
 */
int rtsp_streaming_start(struct stream_t *stream)
{
    
    struct rtsp_header_t *rtsp_answer = NULL;
    struct rmff_header_t *rmff_header;
    int header_size;
  
    char *answer = NULL; /* temporary pointer */
    char *server = NULL;
    char *redirected = NULL;

    stream->stream_ctrl->status = streaming_handshaking;


    if(stream->stream_ctrl->url->protocol_type == RTSP_WMS) {
	server = strdup("WMServer");
    }
    else { /* still don't know which rtsp protocol, or REAL specified (need challenge) */
	
	rtsp_answer = rtsp_connect(stream);
	
	if(rtsp_answer == NULL) {
	    display(MSDL_ERR,"rtsp connection failed\n");
	    
	    goto failed;
	}
	
	/* get server name */
	if((answer = rtsp_get_field(rtsp_answer,"Server")) != NULL) {
	    while(*answer == ' ') answer++;  /* skip space */
	    server = strdup(answer);
	}
	else {
	    if((rtsp_get_field(rtsp_answer,"RealChallenge1")) != NULL) {
		server = strdup("Real");
	    }
	    else {
		server = strdup("Unknown Server");
	    }
	}
    }
  
    
    /* real or helix server (supported) */
    if(strstr(server,"Real") || strstr(server,"Helix")) {
	
	/* this is the function to do almost all setup process */
	rmff_header = real_setup_and_get_header(stream);
    
	if(!rmff_header) { /*  NOT OK  */
	    if((answer = rtsp_get_field(rtsp_answer,"Location")) != NULL) {
		/* redirected */
		while(*answer == ' ') answer++;
	
		redirected = strdup(answer);
		display(MSDL_NOR,"redirected to %s\n",redirected);
		/*
		  TODO redirection support
		  and I have never seen example...
		*/
		goto failed;
	    }
	    else {
		display(MSDL_ERR,"real/helix connection not established\n");
		goto failed;
	    }
	}
    
	/* OK, connection established */
    
	
	header_size = rmff_dump_header(rmff_header,
				       stream->stream_ctrl->write_buffer);
	stream->stream_ctrl->write_data_len = header_size;
    
	display(MSDL_VER,"rmff_header_size = %d\n",header_size);
	stream->stream_ctrl->rtsp_ctrl->rmff_header = rmff_header;
	
	/* set file size to download */
	stream->stream_ctrl->file_size = rmff_header->data->size;
	/* for free() ing */
	stream->stream_ctrl->rtsp_ctrl->rtsp_protocol = RTSP_REAL_PROTOCOL;
    
	stream->stream_ctrl->protocol = RTSP_REAL;

    }
    else if(strstr(server,"WMServer")) {
	
	/* try rtsp-wms */
	struct asf_headerinfo_t *asf_headerinfo;

	if(stream->sock) {
	    close(stream->sock);
	}
    
	free_rtsp_ctrl_t(stream->stream_ctrl->rtsp_ctrl);
	stream->stream_ctrl->rtsp_ctrl = new_rtsp_ctrl_t();
	stream->stream_ctrl->rtsp_ctrl->get_media_packet =
	    wmserver_rtp_get_media_packet;
    
	if(stream->stream_ctrl->url->port == 0) {
	    stream->stream_ctrl->url->port = RTSP_PORT; /* RTSP default port */
	}
    
	stream->stream_ctrl->rtsp_ctrl->server = NULL;
	stream->stream_ctrl->rtsp_ctrl->cseq = 1;
	stream->stream_ctrl->rtsp_ctrl->server_state = 0;
	stream->stream_ctrl->rtsp_ctrl->server_caps  = 0;
	stream->stream_ctrl->rtsp_ctrl->session = NULL;

	/* make mrl */
	rtsp_make_mrl(stream->stream_ctrl->rtsp_ctrl,stream->stream_ctrl->url);
    
	stream->sock = server_connect(stream->stream_ctrl->url->hostname,
				      stream->stream_ctrl->url->port);
	if(stream->sock < 0) { /* couldn't connect for some reason. */
	    stream->stream_ctrl->status = streaming_other_protocol;
	    stream->stream_ctrl->retry_protocol = MMST;	    
	    goto failed;
	}

    
	asf_headerinfo = wmserver_setup_and_get_header(stream);

	if(!asf_headerinfo) {
	    display(MSDL_ERR,"rtsp-wms connection not established\n");
	    display(MSDL_ERR,"server probably does not accept rtsp\n"
		    "retry using mmst protocol\n",server);
	    
	    stream->stream_ctrl->status = streaming_other_protocol;
	    stream->stream_ctrl->retry_protocol = MMST;	    

	    goto failed;
	}
	/* for free()ing */
	stream->stream_ctrl->rtsp_ctrl->rtsp_protocol = RTSP_WMS_PROTOCOL;
	
	/* for msdl function to know protocol */
	stream->stream_ctrl->protocol = RTSP_WMS;
    
    }
    else { /* unsupported servers */
	display(MSDL_ERR,"server type [%s] not supported\n",server);
	
	stream->stream_ctrl->rtsp_ctrl->rtsp_protocol = RTSP_UNKNOWN_PROTOCOL;
	goto failed;
    }
  
  
    if(server) free(server);
    if(rtsp_answer) free_rtsp_header(rtsp_answer);
    if(redirected) free(redirected);
  
  
    stream->stream_ctrl->status = streaming_downloading;

    return 1; /* success */
  
  failed:
    if(server)      free(server);
    if(rtsp_answer) free_rtsp_header(rtsp_answer);
    if(redirected)  free(redirected);
    return -1;
}




/*
 * new_rtsp_header : create new RTSP header
 */
struct rtsp_header_t *new_rtsp_header(void)
{
    struct rtsp_header_t *hdr;
    hdr = (struct rtsp_header_t *)xmalloc(sizeof(struct rtsp_header_t));
    return hdr;
}


/*
 * free RTSP header
 */
void free_rtsp_header(struct rtsp_header_t *h)
{
    struct rtsp_field_t *field, *prev;
  
    if(h == NULL) return; // rtsp_hdr is not malloc()ed yet.

    if(h->protocol) free(h->protocol);
    if(h->method) free(h->method);
    if(h->uri) free(h->uri);
    if(h->reason_phrase) free(h->reason_phrase);
    if(h->field_search) free(h->field_search);
    if(h->buffer) free(h->buffer);

    field = h->first_field;

    for(field = h->first_field; field ; ) {
	if(field->field_name)
	    free(field->field_name);
	prev = field;
	field = field->next;
	free(prev);
    }
    free(h);
}


/*
 * returns first filed with 'field name'.
 * rtsp_get_next_field will get next field with 'filed_name'
 */
char *rtsp_get_field(struct rtsp_header_t *rtsp_hdr, const char *field_name)
{
    if(!rtsp_hdr || !field_name) return NULL;
  
    rtsp_hdr->field_search_pos = rtsp_hdr->first_field;
    rtsp_hdr->field_search = (char *)xrealloc(rtsp_hdr->field_search,
					      strlen(field_name) + 1);
    /* copy field name first. */
    strcpy(rtsp_hdr->field_search, field_name);

    /* get first field value with this field_name. */
    return rtsp_get_next_field(rtsp_hdr);
}



/*
 * return field string after "rtsp_hdr->field_search".
 *        NULL if not found.
 */
char *rtsp_get_next_field(struct rtsp_header_t *rtsp_hdr)
{
    char *ptr;
    struct rtsp_field_t *field;

    if(!rtsp_hdr) return NULL;
    field = rtsp_hdr->field_search_pos;
  
    while(field) {
	ptr = strstr(field->field_name,":");
	if(ptr == NULL) return NULL; // the header is not valid...
	if(!strnicmp(field->field_name,rtsp_hdr->field_search,
			ptr - field->field_name)) { // found field!!!!
	    ptr++; // skip colon.
	    while(*ptr == ' ') ptr++; // skip %20
	    rtsp_hdr->field_search_pos = field->next; // points to next field.
	    return ptr; // return the string without filed name!!
	}
	field = field->next;
    }
    return NULL; // NOT FOUND
}



/*
 * rtsp_set_field : make new field and attach it to last_field->next.
 */
void rtsp_set_field(struct rtsp_header_t *rtsp_hdr, const char *field_name)
{
    struct rtsp_field_t *new_field;
  
    if(rtsp_hdr == NULL || field_name == NULL) return;
  
    new_field = xmalloc(sizeof(struct rtsp_field_t));

    new_field->next = NULL;
    new_field->field_name = xmalloc(strlen(field_name) + 1);
    strcpy(new_field->field_name, field_name);
  
    if(rtsp_hdr->last_field == NULL) {
	rtsp_hdr->first_field = new_field; // this was first filed!!
    }
    else {
	rtsp_hdr->last_field->next = new_field; // attach to last.
    }
    rtsp_hdr->last_field = new_field;
    rtsp_hdr->field_nb++;
}



void rtsp_set_uri(struct rtsp_header_t *rtsp_hdr,const char *uri)
{
    if(rtsp_hdr == NULL || uri == NULL) return;
    rtsp_hdr->uri = xmalloc(strlen(uri) + 1);
    strcpy(rtsp_hdr->uri,uri);
}



void rtsp_set_method(struct rtsp_header_t *rtsp_hdr, const char *method)
{
    if(rtsp_hdr == NULL || method == NULL) return;

    rtsp_hdr->method = xmalloc(strlen(method) + 1);
    strcpy(rtsp_hdr->method,method);
}



/*
 * wrappers for each request
 */
int rtsp_200ok(struct stream_t *stream,int cseq,char *session)
{
    int ret;
    int buflen = 100 + strlen(rtsp_protocol_version) + strlen(session);
    char *buf = xmalloc(buflen + 1);
  
    _snprintf(buf,buflen,
	     "%s 200 OK\r\n"
	     "Cseq: %u\r\n"
	     "Session: %s\r\n"
	     "\r\n",
	     rtsp_protocol_version,
	     cseq,
	     session);
  
    ret = xsend(stream->sock,buf,strlen(buf));
    display(MSDL_DBG,"=send 200 OK============\n%s\n=================\n",buf);
  
    free(buf);
    return ret;
}



/* OPTIONS */
char *rtsp_request_options(struct rtsp_header_t *rtsp_hdr,struct URL_t *url,
			   char *what)
{
    rtsp_set_method(rtsp_hdr,"OPTIONS");
  
    if(what) { /* use what as uri */
	rtsp_set_uri(rtsp_hdr,what);
    }

    else {
	rtsp_hdr->uri = (char *)xmalloc(strlen(url->hostname) + 16);
	sprintf(rtsp_hdr->uri,"rtsp://%s:%i",url->hostname,url->port);
    }
    return (rtsp_build_request(rtsp_hdr));
}


/* DESCRIBE */
char *rtsp_request_describe(struct rtsp_header_t *rtsp_hdr,struct URL_t *url,
			    char *what)
{
    rtsp_set_method(rtsp_hdr,"DESCRIBE");
  
    if(what) {
	rtsp_set_uri(rtsp_hdr,what);
    }
    else {
	rtsp_hdr->uri = (char *)xmalloc(strlen(url->hostname) + 
					strlen(url->filepath) + 16);
	sprintf(rtsp_hdr->uri,"rtsp://%s:%i%s",
		url->hostname,url->port,url->filepath);
    }
    return (rtsp_build_request(rtsp_hdr));
}


/* SETUP */
char *rtsp_request_setup(struct rtsp_header_t *rtsp_hdr,struct URL_t *url,
			 char *what)
{
    rtsp_set_method(rtsp_hdr,"SETUP");
  
    if(what) {
	rtsp_set_uri(rtsp_hdr,what);
    }
  
    return (rtsp_build_request(rtsp_hdr));
}



/* SET_PARAMETER */
char *rtsp_request_set_parameter(struct rtsp_header_t *rtsp_hdr,
				 struct URL_t *url,
				 char *what)
{
    rtsp_set_method(rtsp_hdr,"SET_PARAMETER");
  
    if(what) {
	rtsp_set_uri(rtsp_hdr,what);
    }
    else {
	rtsp_hdr->uri = (char *)xmalloc(strlen(url->hostname) + 
					strlen(url->filepath) + 16);
	sprintf(rtsp_hdr->uri,"rtsp://%s:%i%s",
		url->hostname,url->port,url->filepath);
    }
    return (rtsp_build_request(rtsp_hdr));
}


/* GET_PARAMETER */
char *rtsp_request_get_parameter(struct rtsp_header_t *rtsp_hdr,
				 struct URL_t *url,
				 char *what)
{
    rtsp_set_method(rtsp_hdr,"GET_PARAMETER");
  
    if(what) {
	rtsp_set_uri(rtsp_hdr,what);
    }
    else {
	rtsp_hdr->uri = (char *)xmalloc(strlen(url->hostname) + 
					strlen(url->filepath) + 16);
	sprintf(rtsp_hdr->uri,"rtsp://%s:%i%s",
		url->hostname,url->port,url->filepath);
    }
    return (rtsp_build_request(rtsp_hdr));
}


/* PLAY */
char *rtsp_request_play(struct rtsp_header_t *rtsp_hdr,
			struct URL_t *url,
			char *what)
{
    rtsp_set_method(rtsp_hdr,"PLAY");
  
    if(what) {
	rtsp_set_uri(rtsp_hdr,what);
    }
    else {
	rtsp_hdr->uri = (char *)xmalloc(strlen(url->hostname) + 
					strlen(url->filepath) + 16);
	sprintf(rtsp_hdr->uri,"rtsp://%s:%i%s",
		url->hostname,url->port,url->filepath);
    }
    return (rtsp_build_request(rtsp_hdr));
}


char *rtsp_request_teardown(struct rtsp_header_t *rtsp_hdr,struct URL_t *url,
			    char *what)
{
    rtsp_set_method(rtsp_hdr,"TEARDOWN");
  
    if(what) {
	rtsp_set_uri(rtsp_hdr,what);
    }
    else {
	rtsp_hdr->uri = (char *)xmalloc(strlen(url->hostname) + 
					strlen(url->filepath) + 16);
	sprintf(rtsp_hdr->uri,"rtsp://%s:%i%s",
		url->hostname,url->port,url->filepath);
    }
  
    return (rtsp_build_request(rtsp_hdr));
}



/*
 * make complete rtsp_header_t.
 * firstline is first line, such as OPTIONS, DESCRIBE, etc... (see rtsp)
 * the request string goes to rtsp_hdr->buffer
 */
char *rtsp_build_request(struct rtsp_header_t *rtsp_hdr)
{
    char *ptr;
    int len = 0;
    struct rtsp_field_t *field;

    if(!rtsp_hdr) return NULL;
    if(!rtsp_hdr->method) return NULL;
    if(!rtsp_hdr->uri) return NULL;
  
    /*
     * culculate the request length.
     */
    len = strlen(rtsp_hdr->method) + strlen(rtsp_hdr->uri) +
	strlen(rtsp_protocol_version) + 2 + 2 ; /* 3 is for spaces between*/
  
    // fields
    field = rtsp_hdr->first_field;
    while(field) {
	len += strlen(field->field_name) + 2;
	field = field->next;
    }
    // CRLF at the end
    len += 2;
    // request body
    if(rtsp_hdr->body) {
	len += rtsp_hdr->body_len;
    }
  
    if(rtsp_hdr->buffer) {
	free(rtsp_hdr->buffer);
	rtsp_hdr->buffer = NULL;
    }
 
    if(rtsp_hdr->body_len)
	len += rtsp_hdr->body_len;
  
    rtsp_hdr->buffer_len = len;
    rtsp_hdr->buffer = xmalloc(len + 1); // 1 for '\0'
  
    /* 
     * build the request
     */
    ptr = rtsp_hdr->buffer;

    ptr += sprintf(ptr,"%s %s %s\r\n",rtsp_hdr->method,rtsp_hdr->uri,
		   rtsp_protocol_version);
  
    field = rtsp_hdr->first_field;
    // fields
    while(field != NULL) {
	ptr += sprintf(ptr,"%s\r\n",field->field_name);
	field = field->next;
    }
    ptr += sprintf(ptr,"\r\n");
  
    if(rtsp_hdr->body != NULL) {
	memcpy(ptr,rtsp_hdr->body,rtsp_hdr->body_len);
    }
  
  
    return rtsp_hdr->buffer;
}



/*
 * judges if rtsp_hdr is complete RTSP header, or still needs some parts.
 *    return value :       0 ... NOT complete
 *                         1 ... COMPLETE!!
 */
int rtsp_is_entire_header(struct rtsp_header_t *rtsp_hdr)
{
    if(rtsp_hdr == NULL) return 0; /* unlikely, but error */
  
    if(rtsp_hdr->buffer == NULL) return 0;   /* nothing received. */

    if(strstr(rtsp_hdr->buffer, "\r\n\r\n") == NULL &&        /* M$ style   */
       strstr(rtsp_hdr->buffer, "\n\n") == NULL ) return 0;   /* *nix style */
  
    return 1;
}


static struct rtsp_ctrl_t *new_rtsp_ctrl_t(void)
{
    struct rtsp_ctrl_t *ctrlt = 
	(struct rtsp_ctrl_t *)xmalloc(sizeof(struct rtsp_ctrl_t));

    /* allocate other protocol controllers in
       each protocol initiation functions,
       such as rmff_header or asf_headerinfo...
    */
  
    ctrlt->prev_ts = -1;
    ctrlt->prev_stream_number = -1;
  
    return ctrlt;
}


static void free_rtsp_ctrl_t(struct rtsp_ctrl_t *ctrlt)
{
    if(ctrlt->rtsp_protocol == RTSP_REAL_PROTOCOL) {
	if(ctrlt->rmff_header) {
	    free_rmff_header_t(ctrlt->rmff_header);
	}
    }
    else if(ctrlt->rtsp_protocol == RTSP_WMS_PROTOCOL) {
	if(ctrlt->asf_headerinfo) {
	    free_asf_headerinfo_t(ctrlt->asf_headerinfo);
	}
    }
  
    if(ctrlt->mrl) free(ctrlt->mrl);
    if(ctrlt->server) free(ctrlt->server);
    if(ctrlt->session) free(ctrlt->session);
    if(ctrlt->challenge) free(ctrlt->challenge);
  
    free(ctrlt);
}


void rtsp_streaming_close(struct stream_t *stream)
{

    /* TEARDOWN */
    if(stream->stream_ctrl->status == streaming_downloading || 
       stream->stream_ctrl->status == streaming_finished) {
	struct rtsp_header_t *rtsp_hdr;
	int ret;
	/* send teardown request to make server happier. */
	rtsp_hdr = new_rtsp_header();
	rtsp_set_standard_fields(stream->stream_ctrl->rtsp_ctrl,rtsp_hdr);
	rtsp_request_teardown(rtsp_hdr,stream->stream_ctrl->url,NULL);
	ret = rtsp_send_request(stream,rtsp_hdr);
	free_rtsp_header(rtsp_hdr);
    }
  
    if(stream->sock > 0) { /* valid socket --> close */
	close(stream->sock);
    }
  
    free_rtsp_ctrl_t(stream->stream_ctrl->rtsp_ctrl);
    free_stream_ctrl_t(stream->stream_ctrl);
    free_stream_t(stream);
}


struct stream_t *rtsp_streaming_init(struct URL_t *url)
{
    struct stream_t *stream = new_stream_t();
    stream->stream_ctrl = new_stream_ctrl_t(url);
    stream->stream_ctrl->rtsp_ctrl = new_rtsp_ctrl_t();
  
    stream->start = rtsp_streaming_start;
    stream->read  = rtsp_streaming_read;
    stream->close = rtsp_streaming_close;
  
    return stream;
}
